# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import os
import sys
import requests
from subprocess import run

from ..environment import BaseEnvironment
from ...general import _pkg_root
try:
    from xaux import FsPath  # TODO: once xaux is in Xsuite keep only this
except (ImportError, ModuleNotFoundError):
    from ...xaux import FsPath


class Geant4Environment(BaseEnvironment):
    _read_only_paths = {'bdsim': 0, 'geant4': 0}

    def __init__(self):
        super().__init__()
        self._in_constructor = True
        self._geant4 = None
        self._bdsim = None
        self._in_constructor = False
        self._geant4_sourced = False
        self._bdsim_sourced = False
        try:
            cmd = run(['which', 'geant4-config'], capture_output=True)
        except FileNotFoundError:
            pass
        else:
            if cmd.returncode == 0:
                path = FsPath(cmd.stdout.decode().strip())
                if path.exists():
                    self._geant4 = path
                    self._geant4_sourced = True
        try:
            cmd = run(['which', 'bdsim'], capture_output=True)
        except FileNotFoundError:
            pass
        else:
            if cmd.returncode == 0:
                path = FsPath(cmd.stdout.decode().strip())
                if path.exists():
                    self._bdsim = path
                    self._bdsim_sourced = True

    @property
    def compiled(self):
        if self.geant4 is None or self.bdsim is None:
            return False
        try:
            from g4interface import XtrackInterface
            return True
        except (ModuleNotFoundError, ImportError) as error:
            return False

    @property
    def ready(self):
        return super().ready and self._geant4_sourced and self._bdsim_sourced

    def compile(self, verbose=True, verbose_compile_output=False):
        # Check all dependencies
        self.assert_installed('make', verbose=verbose)
        self.assert_installed('cmake', verbose=verbose)
        self.assert_gcc_installed(verbose=verbose)
        self.assert_gxx_installed(verbose=verbose)
        self.assert_geant4_installed()
        self.assert_bdsim_installed()
        bdsim_version = self.get_bdsim_version()
        if verbose:
            print(f"BDSIM version: {bdsim_version}")

        # Check pybind11 is installed
        try:
            import pybind11 # noqa F401
        except (ModuleNotFoundError, ImportError):
            pybind_repo = 'https://github.com/pybind/pybind11.git'
            try:
                not_found = requests.get(pybind_repo).status_code == 404
            except (requests.exceptions.HTTPError, requests.exceptions.ConnectionError):
                not_found = True
            if not_found:
                raise RuntimeError("Could not find pybind11! Please pip install "
                                  f"pybind11, or make sure {pybind_repo} is accessible.")

        # Copy the C source files to a temporary directory and compile it
        dest = (self.temp_dir / 'xcoll_bdsim').resolve()
        dest.mkdir(parents=True, exist_ok=True)
        for path in (_pkg_root / 'scattering_routines' / 'geant4' / 'scattering_src').glob('*'):
            if path.name == '.git' or path.name == 'docs' or path.name == 'tests' \
            or path.name == 'samples':
                continue
            else:
                FsPath(path).copy_to(dest, method='mount')
        cwd = FsPath.cwd()
        os.chdir(dest)
        self._adapt_source_to_bdsim_version(bdsim_version, verbose)

        # Configure
        ctab = '    '
        cmd = run(['cmake', '-S', '.', '-B', 'build'], capture_output=True)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            os.chdir(cwd)
            raise RuntimeError(f"Failed to build Xcoll-BDSIM interface!\nError given is:\n{stderr}")
        if verbose_compile_output:
            print()
            print("CMake: Configuring")
            print(ctab + cmd.stdout.decode('UTF-8').strip().replace('\n', f'\n{ctab}'))
            if cmd.stderr:
                print(ctab + cmd.stderr.decode('UTF-8').strip().replace('\n', f'\n{ctab}'))
            print()

        # Build
        cmd = run(['cmake', '--build', 'build'], capture_output=True)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            os.chdir(cwd)
            raise RuntimeError(f"Failed to compile Xcoll-BDSIM interface!\nError given is:\n{stderr}")
        if verbose_compile_output:
            print("CMake: Building")
            print(ctab + cmd.stdout.decode('UTF-8').strip().replace('\n', f'\n{ctab}'))
            if cmd.stderr:
                print(ctab + cmd.stderr.decode('UTF-8').strip().replace('\n', f'\n{ctab}'))
            print()
        if verbose:
            print("Compiled Xcoll-BDSIM interface successfully.")

        # Collect the compiled shared library
        so = list((dest / 'build').glob('g4interface.*so'))
        if len(so) > 1:
            raise RuntimeError(f"Compiled into multiple g4interface shared libraries!")
        if len(so) == 0:
            raise RuntimeError(f"Failed Xcoll-BDSIM compilation! No shared library found in "
                             + f"{self.data_dir.as_posix()}!")
        so = FsPath(so[0])
        so.move_to(self.data_dir / so.name)
        if verbose:
            print(f"Created Xcoll-BDSIM shared library in {self.data_dir / so.name}.")
        # Clean up the temporary directory
        self.temp_dir = None
        os.chdir(cwd)

    def assert_geant4_installed(self):
        if self.geant4 is None:
            raise RuntimeError("Could not find Geant4 installation! Please install Geant4.")
        if not self._geant4_sourced:
            raise RuntimeError(f"Geant4 installation found in {self.geant4} "
                               f"but not active! Please source environment.")

    def assert_bdsim_installed(self):
        if self.bdsim is None:
            raise RuntimeError("Could not find BDSIM installation! Please install BDSIM.")
        if not self._bdsim_sourced:
            raise RuntimeError(f"BDSIM installation found in {self.bdsim} "
                               f"but not active! Please source environment.")

    def assert_environment_ready(self):
        self.assert_geant4_installed()
        self.assert_bdsim_installed()
        super().assert_environment_ready()


    def get_bdsim_version(self):
        cmd = run(['bdsim', '--version'], capture_output=True)
        if cmd.returncode != 0:
            stderr = cmd.stderr.decode('UTF-8').strip()
            raise RuntimeError(f"Failed to run 'bdsim --version'!\nError given is:\n{stderr}")
        return cmd.stdout.decode('UTF-8').strip()

    def bdsim_older_than(self, bdsim_version=None, compare_version='1.7.7.develop'):
        if bdsim_version is None:
            bdsim_version = self.get_bdsim_version()
        n_ver = sum([10**(3*(2-i))*int(j) for i, j in enumerate(bdsim_version.strip().split('.')[:3])])
        if 'develop' in bdsim_version:
            n_ver += 0.5
        n_ver_cmp = sum([10**(3*(2-i))*int(j) for i, j in enumerate(compare_version.strip().split('.')[:3])])
        if 'develop' in compare_version:
            n_ver_cmp += 0.5
        return n_ver < n_ver_cmp


    def _adapt_source_to_bdsim_version(self, bdsim_version, verbose):
        version_mark = '// BDSIM >= '
        adapted_any = False

        # Adapt BDSXtrackInterface.hh
        adapted = False
        with open('BDSXtrackInterface.hh', 'r') as file:
            filedata = file.read()
        if self.bdsim_older_than(bdsim_version, '1.7.7.develop'):
            filedata = filedata.replace('#include "BDSLinkBunch.hh"', '#include "BDSBunchSixTrackLink.hh"')
            filedata = filedata.replace('BDSLinkBunch* stp = nullptr;', 'BDSBunchSixTrackLink* stp = nullptr;')
            adapted = True
        new_filedata = []
        for line in filedata.split('\n'):
            if version_mark in line:
                if self.bdsim_older_than(bdsim_version, line.split(version_mark)[1]):
                    adapted = True
                    continue
            new_filedata.append(line)
        if adapted:
            adapted_any = True
            with open('BDSXtrackInterface.hh', 'w') as file:
                file.write('\n'.join(new_filedata))

        # Adapt BDSXtrackInterface.cpp
        adapted = False
        with open('BDSXtrackInterface.cpp', 'r') as file:
            filedata = file.read()
        if self.bdsim_older_than(bdsim_version, '1.7.7.develop'):
            filedata = filedata.replace('stp = new BDSLinkBunch();', 'stp = new BDSBunchSixTrackLink();')
            adapted = True
        new_filedata = []
        for line in filedata.split('\n'):
            if version_mark in line:
                if self.bdsim_older_than(bdsim_version, line.split(version_mark)[1]):
                    adapted = True
                    continue
            new_filedata.append(line)
        if adapted:
            adapted_any = True
            with open('BDSXtrackInterface.cpp', 'w') as file:
                file.write('\n'.join(new_filedata))

        if verbose and adapted_any:
                print("Adapted source code to BDSIM version.")



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
from warnings import warn

import xobjects as xo
import xtrack as xt


class BlowUp(xt.BeamElement):
    _xofields = {
        '_plane':               xo.Int8,
        'start_at_turn':        xo.Int64,
        'stop_at_turn':         xo.Int64,
        'use_individual_kicks': xo.Int8,
        '_max_kick':            xo.Float64,
        '_rans':                xo.Float64[:],
        '_calibration':         xo.Float64,
        '_active':              xo.Int8,
    }

    _isthick = False
    needs_rng = True
    prototype = "blowup"
    allow_track = True
    iscollective = False
    has_backtrack = False
    behaves_like_drift = False
    allow_rot_and_shift = True
    allow_loss_refinement = False
    name_associated_aperture = None
    skip_in_loss_location_refinement = False

    _noexpr_fields = {'plane', 'name', 'line'}
    _skip_in_to_dict = ['_max_kick', '_plane', '_calibration', '_active']
    _store_in_to_dict = ['amplitude', 'plane', 'calibration', 'beta0',
                         'gamma0', 'name']

    _depends_on = [xt.RandomUniform]
    _extra_c_sources = [
        '#include "xcoll/beam_elements/elements_src/blowup.h"'
    ]


    def __init__(self, **kwargs):
        to_assign = {}
        if '_xobject' not in kwargs:
            start_at_turn = int(kwargs.get('start_at_turn', 0))
            stop_at_turn  = int(kwargs.get('stop_at_turn', start_at_turn+1))
            kwargs['start_at_turn'] = start_at_turn
            kwargs['stop_at_turn'] = stop_at_turn
            rans = np.random.uniform(size=stop_at_turn-start_at_turn)
            kwargs['_rans'] = 2*rans - 1
            if 'plane' in kwargs:
                to_assign['plane'] = kwargs.pop('plane')
            to_assign['calibration'] = kwargs.pop('calibration', 1.)
            to_assign['amplitude'] = kwargs.pop('amplitude', 1)
            kwargs['_calibration'] = 1.
            to_assign['_beta0'] = kwargs.pop('beta0', None)
            to_assign['_gamma0'] = kwargs.pop('gamma0', None)
            to_assign['_name'] = kwargs.pop('name', None)
        super().__init__(**kwargs)
        for key, val in to_assign.items():
            setattr(self, key, val)


    @classmethod
    def install(cls, line, name, *, at=None, need_apertures=True,
                aperture=None, s_tol=1.e-6, at_s=None, **kwargs):
        if at_s is not None:
            warn("Warning: `at_s` is deprecated and will be removed in "
                 "the future. Please use `at` instead.", FutureWarning)
            at = at_s
        if name in line.element_names:
            raise ValueError(f"Element {name} already exists in the line as "
                             f"{line[name].__class__.__name__}.")
        self = cls(**kwargs)
        env = line.env
        env.elements[name] = self
        insertions = []
        insertions.append(env.place(name, at=at))
        self._name = name
        if need_apertures:
            if aperture is not None:
                aper_name = f'{name}_aper'
                env.elements[aper_name] = aperture
                self.name_associated_aperture = aper_name
                env.new(f'{aper_name}_upstream', aper_name, mode='replica')
                env.new(f'{aper_name}_downstream', aper_name, mode='replica')
            else:
                idx = line.element_names.index(name)
                while True:
                    if xt.line._is_aperture(line.elements[idx], line):
                        aper_upstream = line.elements[idx].copy()
                        break
                    idx -= 1
                idx = line.element_names.index(name)
                while True:
                    if xt.line._is_aperture(line.elements[idx], line):
                        aper_downstream = line.elements[idx].copy()
                        break
                    idx += 1
                env.elements[f'{name}_aper_upstream'] = aper_upstream
                env.elements[f'{name}_aper_downstream'] = aper_downstream
            insertions.append(env.place(f'{name}_aper_upstream',
                                        at=name+'@start'))
            insertions.append(env.place(f'{name}_aper_downstream',
                                        at=name+'@end'))
        line.insert(insertions, s_tol=s_tol)

        return self


    @property
    def beta0(self):
        return self._beta0

    @property
    def gamma0(self):
        return self._gamma0

    @property
    def name(self):
        return self._name


    @property
    def plane(self):
        if self._plane == 1:
            return 'H'
        elif self._plane == -1:
            return 'V'
        else:
            raise ValueError("The plane of the BlowUp is not recognized.")

    @plane.setter
    def plane(self, val):
        if val.lower() == 'h':
            self._plane = 1
        elif val.lower() == 'v':
            self._plane = -1
        else:
            raise ValueError("The plane of the BlowUp must be either 'H' or "
                             "'V'.")

    @property
    def amplitude(self):
        return (self._max_kick / self.calibration)**2

    @amplitude.setter
    def amplitude(self, val):
        if val< 0:
            raise ValueError("The amplitude cannot be negative!")
        self._max_kick = np.sqrt(val) * self.calibration

    @property
    def calibration(self):
        return self._calibration

    @calibration.setter
    def calibration(self, val):
        if val< 0:
            raise ValueError("The calibration cannot be negative!")
        previous_amplitude = self.amplitude
        self._calibration = val
        self.amplitude = previous_amplitude


    def calibrate_by_emittance(self, nemitt, *, line=None, twiss=None,
                               name=None, beta0=None, gamma0=None):
        # The emittance gained per turn of blow up can be calculated as
        # follows:
        #     <px^2> = gamma * eps
        #     gamma * eps = <px^2> = <(px0 + dpx)^2>
        # where px0 is the initial momentum and dpx is the kick, and
        # eps0 and eps are the geometric emittance before resp. after
        # the kick, and gamma is the Courant-Snyder parameter. If each
        # particle gets another random kick, we can write:
        #     <(px0 + dpx)^2> = <px0^2> + 2<px0 dpx> + <dpx^2>
        #                     = gamma * eps0 + 2<px0 dpx> + 1/3 dpx_max^2    (variance of uniform distribution is 1/12(b-a)^2)
        # The covariance can be estimated with an upper bound:
        #     0  <=  2<px0 dpx>  <=  2*sqrt(<px0^2> <dpx^2>)
        #                         = 2/sqrt(3) * sqrt(gamma * eps0) * dpx_max
        # So that we finally get for the emittance growth after one turn
        # of blow-up:
        #     deps = (eps-eps0)/eps0 = [dPM^2, dPM^2 + 2dPM]    (dPM = dpx_max/sqrt(3 * gamma * eps0))
        # Or, to get an emittance growth of deps, we need:
        #     dpx_max = [sqrt(3* gamma * eps0 * deps), sqrt(3 * gamma * eps0)*(sqrt(1+deps) -1)]
        # If all particles get the same kick, we can write:
        #     <(px0 + dpx)^2> = <px0^2> + 2<px0> dpx
        # However, now things get a bit more complicated. At the start
        # <px0> is the closed orbit, but when the bunch has had a kick
        # in the previous turn, the bunch is no longer centred around
        # the closed orbit. We can write:
        #     deps = (eps-eps0)/eps0 = 2<px0> dpx/gamma/eps0
        if beta0 is None or gamma0 is None:
            if self._beta0 is not None and self._gamma0 is not None:
                beta0 = self._beta0
                gamma0 = self._gamma0
            elif line is not None and hasattr(line, 'particle_ref') \
            and line.particle_ref is not None:
                beta0 = line.particle_ref.beta0[0]
                gamma0 = line.particle_ref.gamma0[0]
            else:
                raise ValueError("Either a line with a particle_ref, or "
                                 "beta0 and gamma0 must be provided!")
        if twiss is None:
            if line is None:
                raise ValueError("Either `line` or `twiss` must be provided!")
            twiss = line.twiss(reverse=False)
        coord = f"gam{'x' if self.plane == 'H' else 'y'}"
        if name is None:
            name = self.name
        gamma = twiss.rows[name][coord][0]
        # Asumming a Gaussian beam, we will have shifted all beam beyond
        # 5 sigma with an emittance growth of around a factor 60. In
        # practice, a ~50x increase is enough due to the covariance
        # terms. We want to do this slowly, over 1000 turns.
        deps = 50/1000
        gemitt = nemitt / beta0 / gamma0
        self.calibration = np.sqrt(3*gamma*gemitt*deps)

    def activate(self):
        self._active = True

    def deactivate(self):
        self._active = False



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
from warnings import warn

import xobjects as xo
import xtrack as xt

from ..general import _pkg_root


class EmittanceMonitorRecord(xo.Struct):
    count            = xo.Float64[:]
    cached           = xo.Int8[:]
    cached_modes     = xo.Int8[:]
    count            = xo.Float64[:]
    x_sum1           = xo.Float64[:]
    px_sum1          = xo.Float64[:]
    y_sum1           = xo.Float64[:]
    py_sum1          = xo.Float64[:]
    zeta_sum1        = xo.Float64[:]
    pzeta_sum1       = xo.Float64[:]
    x_x_sum2         = xo.Float64[:]
    x_px_sum2        = xo.Float64[:]
    x_y_sum2         = xo.Float64[:]
    x_py_sum2        = xo.Float64[:]
    x_zeta_sum2      = xo.Float64[:]
    x_pzeta_sum2     = xo.Float64[:]
    px_px_sum2       = xo.Float64[:]
    px_y_sum2        = xo.Float64[:]
    px_py_sum2       = xo.Float64[:]
    px_zeta_sum2     = xo.Float64[:]
    px_pzeta_sum2    = xo.Float64[:]
    y_y_sum2         = xo.Float64[:]
    y_py_sum2        = xo.Float64[:]
    y_zeta_sum2      = xo.Float64[:]
    y_pzeta_sum2     = xo.Float64[:]
    py_py_sum2       = xo.Float64[:]
    py_zeta_sum2     = xo.Float64[:]
    py_pzeta_sum2    = xo.Float64[:]
    zeta_zeta_sum2   = xo.Float64[:]
    zeta_pzeta_sum2  = xo.Float64[:]
    pzeta_pzeta_sum2 = xo.Float64[:]


class EmittanceMonitor(xt.BeamElement):
    _xofields={
        'part_id_start':      xo.Int64,
        'part_id_end':        xo.Int64,
        'start_at_turn':      xo.Int64,
        'stop_at_turn':       xo.Int64,
        'frev':               xo.Float64,
        'sampling_frequency': xo.Float64,
        '_index':             xt.RecordIndex,
        'data':               EmittanceMonitorRecord,
        '_plane_selector':    xo.Int8
    }

    behaves_like_drift = True
    allow_loss_refinement = True

    _noexpr_fields   = {'name', 'line'}
    _extra_c_sources = [
        xt._pkg_root.joinpath('headers/atomicadd.h'),
        _pkg_root.joinpath('beam_elements/elements_src/emittance_monitor.h')
    ]

    def __init__(self, **kwargs):
        """
        Monitor to save the normalised beam emittance

        Similar to the BeamSizeMonitor and BeamPositionMonitor, it allows for
        arbitrary sampling rate and can thus not only be used to monitor bunch
        emittance, but also to record coasting beams. See their documentation
        for more information on how to use `frev` and `sampling_frequency`.

        Args:
            num_particles (int, optional): Number of particles to monitor,
                starting from 0. Defaults to -1 which means ALL.
            particle_id_range (tuple, optional): Range of particle ids to
                monitor (start, stop). Stop is exclusive. Defaults to
                (particle_id_start, particle_id_start+num_particles).
            start_at_turn (int): First turn of reference particle (inclusive)
                at which to monitor. Defaults to 0.
            stop_at_turn (int): Last turn of reference particle (exclusive) at
                which to monitor. Defaults to start_at_turn + 1.
            frev (float): Revolution frequency in Hz of circulating beam (used
                to relate turn number to sample index). Defaults to 1.
            sampling_frequency (float): Sampling frequency in Hz. Defaults to 1.
            horizontal (bool): Whether or not to monitor the horizontal plane.
                Defaults to True.
            vertical (bool): Whether or not to monitor the vertical plane.
                Defaults to True.
            longitudinal (bool): Whether or not to monitor the longitudinal plane.
                Defaults to True.
        """
        if '_xobject' not in kwargs:
            if 'particle_id_range' in kwargs:
                assert 'num_particles' not in kwargs
                particle_id_range = kwargs.pop('particle_id_range')
                kwargs['part_id_start'] = int(particle_id_range[0])
                kwargs['part_id_end'] = int(particle_id_range[1])
            else:
                num_particles = int(kwargs.pop('num_particles', -1))
                if num_particles == -1:
                    kwargs['part_id_start'] = 0
                    kwargs['part_id_end'] = -1
                else:
                    kwargs['part_id_start'] = kwargs.pop('particle_id_start', 0)
                    kwargs['part_id_end'] = kwargs['part_id_start'] + num_particles
            kwargs['start_at_turn'] = int(kwargs.get('start_at_turn', 0))
            kwargs['stop_at_turn']  = int(kwargs.get('stop_at_turn', kwargs['start_at_turn']+1))
            kwargs.setdefault('frev', 1.)
            kwargs.setdefault('sampling_frequency', 1.)
            horizontal = kwargs.pop('horizontal', True)
            vertical = kwargs.pop('vertical', True)
            longitudinal = kwargs.pop('longitudinal', True)
            kwargs['_plane_selector']  =   int(horizontal)
            kwargs['_plane_selector'] += 2*int(vertical)
            kwargs['_plane_selector'] += 4*int(longitudinal)
            if "data" not in kwargs:
                # explicitely init with zeros (instead of size only) to have consistent initial values
                size = int(round((kwargs['stop_at_turn'] - kwargs['start_at_turn']) \
                                  * kwargs['sampling_frequency'] / kwargs['frev']))
                size_h = np.zeros(size) if horizontal else np.zeros(1)
                size_v = np.zeros(size) if vertical else np.zeros(1)
                size_l = np.zeros(size) if longitudinal else np.zeros(1)
                kwargs['data'] = {field.name: np.zeros(size) for field in EmittanceMonitorRecord._fields
                                  if 'x' not in field.name and 'y' not in field.name and 'zeta' not in field.name}
                kwargs['data'].update({field.name: size_h for field in EmittanceMonitorRecord._fields if 'x' in field.name})
                kwargs['data'].update({field.name: size_v for field in EmittanceMonitorRecord._fields if 'y' in field.name})
                kwargs['data'].update({field.name: size_l for field in EmittanceMonitorRecord._fields if 'zeta' in field.name})
            suppress_warnings = kwargs.pop('suppress_warnings', False)
        super().__init__(**kwargs)
        self._suppress_warnings = suppress_warnings


    def copy(self, *args, **kwargs):
        """Create a copy of the monitor including its data."""
        line = self.line
        del self._line    # Have to delete to avoid copying all line elements (including monitor itself)
        new_monitor = super().copy(*args, **kwargs)
        self._line = line # Restore
        new_monitor._line = line
        return new_monitor


    @classmethod
    def install(cls, line, name, *, at=None, s_tol=1.e-6, at_s=None, **kwargs):
        if at_s is not None:
            warn("Warning: `at_s` is deprecated and will be removed in "
                 "the future. Please use `at` instead.", FutureWarning)
            at = at_s
        if name in line.element_names:
            raise ValueError(f"Element {name} already exists in the line as {line[name].__class__.__name__}.")
        self = cls(**kwargs)
        line.insert(name, self, at=at, s_tol=s_tol)
        self._name = name
        self._line = line
        return self

    def reset(self):
        """Reset the monitor data (to avoid accumulation after re-tracking)."""
        for field in [f.name for f in EmittanceMonitorRecord._fields]:
            ff = getattr(self.data, field)
            zeros = np.zeros(len(ff), dtype=ff._itemtype._dtype)
            setattr(self.data, field, zeros)

    @property
    def suppress_warnings(self):
        return self._suppress_warnings

    @suppress_warnings.setter
    def suppress_warnings(self, val):
        self._suppress_warnings = bool(val)


    @property
    def gemitt_x(self):
        self._check_horizontal()
        self._calculate()
        return self._gemitt_x

    @property
    def gemitt_y(self):
        self._check_vertical()
        self._calculate()
        return self._gemitt_y

    @property
    def gemitt_zeta(self):
        self._check_longitudinal()
        self._calculate()
        return self._gemitt_zeta

    @property
    def nemitt_x(self):
        self._check_horizontal()
        self._calculate()
        return self._gemitt_x * self.beta0 * self.gamma0

    @property
    def nemitt_y(self):
        self._check_vertical()
        self._calculate()
        return self._gemitt_y * self.beta0 * self.gamma0

    @property
    def nemitt_zeta(self):
        self._check_longitudinal()
        self._calculate()
        return self._gemitt_zeta * self.beta0 * self.gamma0

    @property
    def gemitt_I(self):
        self._check_horizontal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_I

    @property
    def gemitt_II(self):
        self._check_vertical()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_II

    @property
    def gemitt_III(self):
        self._check_longitudinal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_III

    @property
    def nemitt_I(self):
        self._check_horizontal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_I * self.beta0 * self.gamma0

    @property
    def nemitt_II(self):
        self._check_vertical()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_II * self.beta0 * self.gamma0

    @property
    def nemitt_III(self):
        self._check_longitudinal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_III * self.beta0 * self.gamma0


    @property
    def horizontal(self):
        return bool(self._plane_selector % 2)

    @property
    def vertical(self):
        return bool((self._plane_selector >> 1) % 2)

    @property
    def longitudinal(self):
        return bool((self._plane_selector >> 2) % 2)


    @property
    def beta0(self):
        if not hasattr(self, '_beta0'):
            self._beta0 = self.line.particle_ref.beta0[0]
        return self._beta0

    @property
    def gamma0(self):
        if not hasattr(self, '_gamma0'):
            self._gamma0 = self.line.particle_ref.gamma0[0]
        return self._gamma0

    @property
    def turns(self):
        self._calculate()
        return self._turns

    @property
    def name(self):
        if not hasattr(self, '_name'):
            raise ValueError("Name not set! Install the monitor using xc.EmittanceMonitor.install() "
                             "or manually set the name after installation.")
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    @property
    def line(self):
        if not hasattr(self, '_line'):
            raise ValueError("Line not set! Install the monitor using xc.EmittanceMonitor.install() "
                             "or manually set the line after installation.")
        return self._line

    @line.setter
    def line(self, val):
        self._line = val


    def _check_horizontal(self):
        if not self.horizontal:
            raise ValueError("Horizontal plane not monitored!")

    def _check_vertical(self):
        if not self.vertical:
            raise ValueError("Vertical plane not monitored!")

    def _check_longitudinal(self):
        if not self.longitudinal:
            raise ValueError("Longitudinal plane not monitored!")


    def _calculate(self):
        # Calculate mean, variance, and std
        N = self.count
        mask = (N > 0) & (self.cached == 0)
        if not np.any(mask):
            return

        N = N[mask]
        self._turns = np.array(range(self.start_at_turn,
                                     self.stop_at_turn))[mask]

        # NaN for zero particles is expected behaviour
        with np.errstate(invalid='ignore'):
            for field in [f.name for f in EmittanceMonitorRecord._fields]:
                if field.endswith('_sum2'):
                    x1, x2 = field[:-5].split('_')
                    ff1 = getattr(self, f'{x1}_sum1')
                    ff1 = ff1[mask] if len(ff1) == len(mask) else ff1
                    mean1 = ff1 / N
                    ff2 = getattr(self, f'{x2}_sum1')
                    ff2 = ff2[mask] if len(ff2) == len(mask) else ff2
                    mean2 = ff2 / N
                    ff = getattr(self, field)
                    ff = ff[mask] if len(ff) == len(mask) else ff
                    variance = ff / (N - 1) - mean1 * mean2 * N / (N - 1)
                    # Assign to attribute (create if does not exist yet):
                    attr_name = f'_{x1}_{x2}_var'
                    arr = getattr(self, attr_name, None)
                    if arr is None:
                        arr = np.full(self.count.shape, -1, dtype=float)
                        setattr(self, attr_name, arr)
                    arr[mask] = variance

        # Set calculated values as cached
        for i in np.arange(len(self.count))[mask]:
            self.data.cached[i] = 1
            self.data.cached_modes[i] = 0

        # Store calculated emittances in attributes (create if needed):
        gemitt_x = np.sqrt(self.x_x_var[mask] * self.px_px_var[mask]
                           - self.x_px_var[mask]**2)
        gemitt_y = np.sqrt(self.y_y_var[mask] * self.py_py_var[mask]
                           - self.y_py_var[mask]**2)
        gemitt_zeta = np.sqrt(self.zeta_zeta_var[mask]
                              * self.pzeta_pzeta_var[mask]
                              - self.zeta_pzeta_var[mask]**2)
        for nn, val in zip(['gemitt_x', 'gemitt_y', 'gemitt_zeta'],
                           [gemitt_x, gemitt_y, gemitt_zeta]):
            arr = getattr(self, f'_{nn}', None)
            if arr is None:
                arr = np.full(self.count.shape, -1, dtype=float)
                setattr(self, f'_{nn}', arr)
            arr[mask] = np.array(val)


    def _calculate_modes(self):
        # Calculate emittance modes
        N = self.count
        mask = (N > 0) & (self.cached_modes == 0)
        if not np.any(mask):
            return

        S = np.array([[ 0., 1., 0., 0., 0., 0.],
                      [-1., 0., 0., 0., 0., 0.],
                      [ 0., 0., 0., 1., 0., 0.],
                      [ 0., 0.,-1., 0., 0., 0.],
                      [ 0., 0., 0., 0., 0., 1.],
                      [ 0., 0., 0., 0.,-1., 0.]])

        # Create arrays for modes in attributes if needed
        for nn in ['gemitt_I', 'gemitt_II', 'gemitt_III']:
            if getattr(self, f'_{nn}', None) is None:
                arr = np.full(N.shape, -1, dtype=float)
                setattr(self, f'_{nn}', arr)

        # Loop over turns
        for i in range(len(N)):
            if not mask[i]:
                continue

            if N[i] < 25:
                # Not enough statistics for a reliable calculation of the modes
                self._gemitt_I[i] = 0
                self._gemitt_II[i] = 0
                self._gemitt_III[i] = 0
                continue

            if self.horizontal:
                block_x = np.array([[self.x_x_var[i],  self.x_px_var[i]],
                                    [self.x_px_var[i], self.px_px_var[i]]])
            else:
                block_x = np.zeros((2, 2))
            if self.vertical:
                block_y = np.array([[self.y_y_var[i],  self.y_py_var[i]],
                                    [self.y_py_var[i], self.py_py_var[i]]])
            else:
                block_y = np.zeros((2, 2))
            if self.longitudinal:
                block_z = np.array([[self.zeta_zeta_var[i],
                                     self.zeta_pzeta_var[i]],
                                    [self.zeta_pzeta_var[i],
                                     self.pzeta_pzeta_var[i]]])
            else:
                block_z = np.zeros((2, 2))
            if self.horizontal and self.vertical:
                block_xy = np.array([[self.x_y_var[i],  self.x_py_var[i]],
                                     [self.px_y_var[i], self.px_py_var[i]]])
            else:
                block_xy = np.zeros((2, 2))
            if self.horizontal and self.longitudinal:
                block_xz = np.array([[self.x_zeta_var[i],
                                      self.x_pzeta_var[i]],
                                     [self.px_zeta_var[i],
                                      self.px_pzeta_var[i]]])
            else:
                block_xz = np.zeros((2, 2))
            if self.vertical and self.longitudinal:
                block_yz = np.array([[self.y_zeta_var[i],
                                      self.y_pzeta_var[i]],
                                     [self.py_zeta_var[i],
                                      self.py_pzeta_var[i]]])
            else:
                block_yz = np.zeros((2, 2))

            covariance_S = np.dot(
                        np.block([[block_x,    block_xy,   block_xz],
                                  [block_xy.T, block_y,    block_yz],
                                  [block_xz.T, block_yz.T, block_z]]),
                        S)

            # Check for all zero matrix -> zero emittance
            if np.all(covariance_S < 1E-16):
                self._gemitt_I[i] = 0
                self._gemitt_II[i] = 0
                self._gemitt_III[i] = 0
                continue

            cond_number = np.linalg.cond(covariance_S)
            if cond_number > 1e10 and not self.suppress_warnings:
                print(f"Warning: High condition number at time step {i}: "
                    + f"{cond_number}.\n{N[i]} particles logged.")

            rank = np.linalg.matrix_rank(covariance_S)
            expected_rank = int(self.horizontal) + int(self.vertical) + int(self.longitudinal)
            if rank < expected_rank and not self.suppress_warnings:
                print(f"Warning: Matrix is rank deficient at time step {i}: "
                    + f"rank {rank} instead of expected {len(covariance_S)}.\n"
                    + f"{N[i]} particles logged.")

            from xtrack.linear_normal_form import compute_linear_normal_form
            _, _, _, eigenvalues = compute_linear_normal_form(covariance_S)
            self._gemitt_I[i] = eigenvalues[0].imag
            self._gemitt_II[i] = eigenvalues[1].imag
            self._gemitt_III[i] = eigenvalues[2].imag

            # Set calculated modes as cached
            self.data.cached_modes[i] = 1


    def __getattr__(self, attr):
        if attr in [f.name for f in EmittanceMonitorRecord._fields]:
            return getattr(self.data, attr).to_nparray()

        elif attr in self.__class__._xofields:
            return super().__getattr(attr)

        else:
            if attr.startswith('_'):
                raise AttributeError(f"Attribute {attr} not set!")

            if attr.endswith('_var'):
                self._calculate()
                return getattr(self, f'_{attr}')

            else:
                raise AttributeError(f"EmittanceMonitor has no attribute '{attr}'")



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import numpy as np
from warnings import warn

import xobjects as xo
import xtrack as xt

from ..general import _pkg_root
from .base import InvalidXcoll


class BlowUp(InvalidXcoll):
    _xofields = {
        '_plane':               xo.Int8,
        'start_at_turn':        xo.Int64,
        'stop_at_turn':         xo.Int64,
        'use_individual_kicks': xo.Int8,
        '_max_kick':            xo.Float64,
        '_rans':                xo.Float64[:],
        '_calibration':         xo.Float64,
        '_active':              xo.Int8,
    }

    isthick = False
    behaves_like_drift = False
    allow_track = True
    needs_rng = True
    skip_in_loss_location_refinement = True
    allow_loss_refinement = False

    _noexpr_fields    = {'plane', 'name', 'line'}
    _skip_in_to_dict  = ['_max_kick', '_plane', '_calibration', '_active']
    _store_in_to_dict = ['amplitude', 'plane', 'calibration']

    _depends_on = [InvalidXcoll, xt.RandomUniform]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','blowup.h')
    ]

    def __init__(self, **kwargs):
        to_assign = {}
        if '_xobject' not in kwargs:
            start_at_turn = int(kwargs.get('start_at_turn', 0))
            stop_at_turn  = int(kwargs.get('stop_at_turn', start_at_turn+1))
            kwargs['start_at_turn'] = start_at_turn
            kwargs['stop_at_turn']  = stop_at_turn
            kwargs['_rans'] = 2*np.random.uniform(size=stop_at_turn-start_at_turn) - 1
            if 'plane' in kwargs:
                to_assign['plane']   = kwargs.pop('plane')
            to_assign['calibration'] = kwargs.pop('calibration', 1.)
            to_assign['amplitude']   = kwargs.pop('amplitude', 1)
            kwargs['_calibration']   = 1.
        super().__init__(**kwargs)
        for key, val in to_assign.items():
            setattr(self, key, val)


    @classmethod
    def install(cls, line, name, *, at=None, need_apertures=True, aperture=None, s_tol=1.e-6, at_s=None, **kwargs):
        if at_s is not None:
            warn("Warning: `at_s` is deprecated and will be removed in "
                 "the future. Please use `at` instead.", FutureWarning)
            at = at_s
        if name in line.element_names:
            raise ValueError(f"Element {name} already exists in the line as {line[name].__class__.__name__}.")
        self = cls(**kwargs)
        env = line.env
        env.elements[name] = self
        insertions = []
        insertions.append(env.place(name, at=at))
        self._name = name
        self._line = line
        if need_apertures:
            if aperture is not None:
                aper_upstream   = aperture.copy()
                aper_downstream = aperture.copy()
            else:
                idx = line.element_names.index(name)
                while True:
                    if xt.line._is_aperture(line.elements[idx], line):
                        aper_upstream = line.elements[idx].copy()
                        break
                    idx -= 1
                idx = line.element_names.index(name)
                while True:
                    if xt.line._is_aperture(line.elements[idx], line):
                        aper_downstream = line.elements[idx].copy()
                        break
                    idx += 1
            env.elements[f'{name}_aper_upstream'] = aper_upstream
            env.elements[f'{name}_aper_downstream'] = aper_downstream
            insertions.append(env.place(f'{name}_aper_upstream', at=name+'@start'))
            insertions.append(env.place(f'{name}_aper_downstream', at=name+'@end'))
        line.insert(insertions, s_tol=s_tol)

        return self

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return xt.Marker(_context=_context, _buffer=_buffer, _offset=_offset)


    @property
    def plane(self):
        if self._plane == 1:
            return 'H'
        elif self._plane == -1:
            return 'V'
        else:
            raise ValueError("The plane of the BlowUp is not recognized.")

    @plane.setter
    def plane(self, val):
        if val.lower() == 'h':
            self._plane = 1
        elif val.lower() == 'v':
            self._plane = -1
        else:
            raise ValueError("The plane of the BlowUp must be either 'H' or 'V'.")

    @property
    def amplitude(self):
        return (self._max_kick / self.calibration)**2

    @amplitude.setter
    def amplitude(self, val):
        if val< 0:
            raise ValueError("The amplitude cannot be negative!")
        self._max_kick = np.sqrt(val) * self.calibration

    @property
    def calibration(self):
        return self._calibration

    @calibration.setter
    def calibration(self, val):
        if val< 0:
            raise ValueError("The calibration cannot be negative!")
        previous_amplitude = self.amplitude
        self._calibration = val
        self.amplitude = previous_amplitude

    @property
    def name(self):
        if not hasattr(self, '_name'):
            raise ValueError("Name not set! Install the blow-up using xc.BlowUp.install() "
                             "or manually set the name after installation.")
        return self._name

    @name.setter
    def name(self, val):
        self._name = val

    @property
    def line(self):
        if not hasattr(self, '_line'):
            raise ValueError("Line not set! Install the blow-up using xc.BlowUp.install() "
                             "or manually set the line after installation.")
        return self._line

    @line.setter
    def line(self, val):
        self._line = val


    def calibrate_by_emittance(self, nemitt, *, twiss=None, beta_gamma_rel=None):
        # The emittance gained per turn of blow up can be calculated as follows:
        #     <px^2> = gamma * eps
        #     gamma * eps = <px^2> = <(px0 + dpx)^2>
        # where px0 is the initial momentum and dpx is the kick, and eps0 and eps
        # are the geometric emittance before resp. after the kick, and gamma is the
        # Courant-Snyder parameter.
        # If each particle gets another random kick, we can write:
        #     <(px0 + dpx)^2> = <px0^2> + 2<px0 dpx> + <dpx^2>
        #                     = gamma * eps0 + 2<px0 dpx> + 1/3 dpx_max^2    (variance of uniform distribution is 1/12(b-a)^2)
        # The covariance can be estimated with an upper bound:
        #     0  <=  2<px0 dpx>  <=  2*sqrt(<px0^2> <dpx^2>)
        #                         = 2/sqrt(3) * sqrt(gamma * eps0) * dpx_max
        # So that we finally get for the emittance growth after one turn of blow-up:
        #     deps = (eps-eps0)/eps0 = [dPM^2, dPM^2 + 2dPM]    (dPM = dpx_max/sqrt(3 * gamma * eps0))
        # Or, to get an emittance growth of deps, we need:
        #     dpx_max = [sqrt(3* gamma * eps0 * deps), sqrt(3 * gamma * eps0)*(sqrt(1+deps) -1)]
        # If all particles get the same kick, we can write:
        #     <(px0 + dpx)^2> = <px0^2> + 2<px0> dpx
        # However, now things get a bit more complicated. At the start <px0> is the closed orbit,
        # but when the bunch has had a kick in the previous turn, the bunch is no longer centred
        # around the closed orbit. We can write:
        #     deps = (eps-eps0)/eps0 = 2<px0> dpx/gamma/eps0
        name = self.name
        line = self.line
        if beta_gamma_rel is None:
            if not hasattr(line, 'particle_ref'):
                raise ValueError("The provided line has no reference particle. Use the argument `beta_gamma_rel`")
            beta_gamma_rel = line.particle_ref.gamma0[0]*line.particle_ref.beta0[0]
        if twiss is None:
            twiss = line.twiss()
        gamma = twiss.rows[name][f"gam{'x' if self.plane == 'H' else 'y'}"][0]
        # Asumming a Gaussian beam, we will have shifted all beam beyond 5 sigma with an emittance growth
        # of around a factor 60. In practive, a ~50x increase is enough due to the covariance terms.
        # We want to do this slowly, over 1000 turns.
        deps = 50/1000
        self.calibration = np.sqrt(3*gamma*nemitt/beta_gamma_rel*deps)

    def activate(self):
        self._active = True

    def deactivate(self):
        self._active = False



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xobjects as xo
import xtrack as xt

from .interaction_types import interactions_src, interaction_names, shortcuts
from ..general import _pkg_root
from ..headers.particle_states import particle_states_src

import numpy as np
import pandas as pd

interaction_names = {kk: vv.replace('_', ' ').title().\
                            replace('Pn ','PN ').replace('Pp ','PP ').replace(' Mcs',' MCS').\
                            replace(' Ch',' CH').replace(' Vr',' VR')
                     for kk, vv in interaction_names.items()}


class InteractionRecord(xt.BeamElement):
    _xofields = {
        '_index':            xt.RecordIndex,
        'at_element':        xo.Int64[:],
        'at_turn':           xo.Int64[:],
        '_inter':            xo.Int64[:],
        'id_before':         xo.Int64[:],
        's_before':          xo.Float64[:],
        'x_before':          xo.Float64[:],
        'px_before':         xo.Float64[:],
        'y_before':          xo.Float64[:],
        'py_before':         xo.Float64[:],
        'zeta_before':       xo.Float64[:],
        'delta_before':      xo.Float64[:],
        'energy_before':     xo.Float64[:],
        'mass_before':       xo.Float64[:],
        'charge_before':     xo.Int64[:],
        'z_before':          xo.Int64[:],
        'a_before':          xo.Int64[:],
        'pdgid_before':      xo.Int64[:],
        'id_after':          xo.Int64[:],
        's_after':           xo.Float64[:],
        'x_after':           xo.Float64[:],
        'px_after':          xo.Float64[:],
        'y_after':           xo.Float64[:],
        'py_after':          xo.Float64[:],
        'zeta_after':        xo.Float64[:],
        'delta_after':       xo.Float64[:],
        'energy_after':      xo.Float64[:],
        'mass_after':        xo.Float64[:],
        'charge_after':      xo.Int64[:],
        'z_after':           xo.Int64[:],
        'a_after':           xo.Int64[:],
        'pdgid_after':       xo.Int64[:],
    }

    allow_track = False

    _extra_c_sources = [
        interactions_src,
        particle_states_src,
        _pkg_root.joinpath('interaction_record','interaction_record_src','interaction_record.h')
    ]


    @classmethod
    def start(cls, *, line=None, elements=None, names=None, record_impacts=None, record_exits=None,
              record_scatterings=None, capacity=1e6, io_buffer=None, coll_ids=None):
        elements, names = _get_xcoll_elements(line, elements, names)
        if len(names) == 0:
            return
        capacity = int(capacity)

        if getattr(line, 'tracker', None) is None \
        or getattr(line.tracker, 'io_buffer', None) is None:
            if io_buffer is None:
                io_buffer = xt.new_io_buffer(capacity=capacity)
        elif io_buffer is not None:
            raise ValueError("Cannot provide io_buffer when tracker already built!")
        else:
            io_buffer = line.tracker.io_buffer
        if capacity > io_buffer.capacity:
            io_buffer.grow(capacity - io_buffer.capacity)
        if record_impacts is None and record_scatterings is None:
            record_impacts = True
            record_scatterings = True
        elif record_impacts is None:
            record_impacts = not record_scatterings
        elif record_scatterings is None:
            record_scatterings = not record_impacts
        if record_exits is None:
            # record_exits defaults to True only if the other two are True
            record_exits = record_impacts and record_scatterings
        assert record_impacts is True or record_impacts is False
        assert record_exits is True or record_exits is False
        assert record_scatterings is True or record_scatterings is False
        for el in elements:
            if not el.record_impacts and not el.record_exits and not el.record_scatterings:
                el.record_impacts = record_impacts
                el.record_exits = record_exits
                el.record_scatterings = record_scatterings
        record = xt.start_internal_logging(io_buffer=io_buffer, capacity=capacity, \
                                           elements=elements)
        record._line = line
        record._io_buffer = io_buffer
        recording_elements = names if len(names) > 0 else elements
        record._recording_elements = recording_elements
        if coll_ids is None:
            if line is None:
                if len(names) > 0:
                    record._coll_ids = {name: idx for idx, name in enumerate(names)}
            else:
                record._coll_ids = {name: line.element_names.index(name) for name in names}
        else:
            assert len(coll_ids) == len(names)
            record._coll_ids = {name: idx for name, idx in zip(names, coll_ids)}
        if hasattr(record, '_coll_ids'):
            record._coll_names = {vv: kk for kk, vv in record._coll_ids.items()}
        return record

    def stop(self, *, elements=False, names=False):
        self.assert_class_init()
        elements, names = _get_xcoll_elements(self.line, elements, names)
        if self.line is not None and self.line.tracker is not None:
            self.line.tracker._check_invalidated()
        xt.stop_internal_logging(elements=elements)
        # Removed the stopped collimators from list of logged elements
        stopping_elements = names if len(names) > 0 else elements
        self._recording_elements = list(set(self._recording_elements) - set(stopping_elements))


    def assert_class_init(self):
        if not hasattr(self, '_io_buffer') or not hasattr(self, '_line') \
        or not hasattr(self, '_recording_elements'):
            raise ValueError("This InteractionRecord has been manually instantiated, "
                           + "hence the expanded API is not available. Use "
                           + "InteractionRecord.start() to initialise with extended API.")

    @property
    def line(self):
        if hasattr(self, '_line'):
            return self._line

    @property
    def io_buffer(self):
        if hasattr(self, '_io_buffer'):
            return self._io_buffer

    @property
    def capacity(self):
        if hasattr(self, '_io_buffer'):
            return self.io_buffer.capacity

    # @capacity.setter
    # def capacity(self, val):
    #     if hasattr(self, '_io_buffer'):
    #         capacity = int(capacity)
    #         if capacity < self.capacity:
    #             raise NotImplementedError("Shrinking of capacity not yet implemented!")
    #         elif capacity == self.capacity:
    #             return
    #         else:
    #             self.io_buffer.grow(capacity - self.capacity)
    #             # TODO: increase capacity of iobuffer AND of fields in record table

    @property
    def recording_elements(self):
        if hasattr(self, '_recording_elements'):
            return self._recording_elements

    @recording_elements.setter
    def recording_elements(self, val):
        self.assert_class_init()
        if val is None:
            val = []
        record_start = _get_xcoll_elements(self.line, val)
        self.stop(set(self.recording_elements) - set(record_start))
        elements = [self.line[name] for name in record_start]
        for el in elements:  # TODO: this should be smarter
            if not el.record_impacts and not el.record_scatterings:
                el.record_impacts = True
                el.record_scatterings = True
        xt.start_internal_logging(io_buffer=self.io_buffer, capacity=self.capacity, \
                                  record=self, elements=elements)
        self._recording_elements = record_start
        # Updating coll IDs: careful to correctly overwrite existing values
        self._coll_ids.update({name: self.line.element_names.index(name) for name in record_start})
        self._coll_names = {vv: kk for kk, vv in self._coll_ids.items()}

    @property
    def interaction_type(self):
        return np.array([interaction_names[inter] for inter in self._inter])

    def _collimator_name(self, element_id):
        if not hasattr(self, '_coll_names'):
            return element_id
        elif element_id not in self._coll_names:
            raise ValueError(f"Element {element_id} not found in list of collimators of this record table! "
                           + f"Did the line change without updating the list in the table?")
        else:
            return self._coll_names[element_id]

    def _collimator_id(self, element_name):
        if not hasattr(self, '_coll_ids'):
            return element_name
        elif element_name not in self._coll_ids:
            raise ValueError(f"Element {element_name} not found in list of collimators of this record table! "
                           + f"Did the line change without updating the list in the table?")
        else:
            return self._coll_ids[element_name]

    def to_pandas(self, frame=None):
        if frame is None:
            frame = 'jaw'
        frame = frame.lower()
        if frame not in ['jaw', 'collimator', 'lab']:
            raise ValueError(f"Invalid frame {frame}. Must be 'jaw', 'collimator', or 'lab'!")
        n_rows = self._index.num_recorded
        coll_header = 'collimator' if hasattr(self, '_coll_names') else 'collimator_id'
        df = pd.DataFrame({
                'turn':              self.at_turn[:n_rows],
                coll_header:         [self._collimator_name(element_id) for element_id in self.at_element[:n_rows]],
                'interaction_type':  [interaction_names[inter] for inter in self._inter[:n_rows]],
                **{
                    f'{val}_{p}': getattr(self, f'{val}_{p}')[:n_rows]
                    for p in ['before', 'after']
                    for val in ['id', 's', 'x', 'px', 'y', 'py', 'zeta', 'delta', 'energy', 'mass', 'charge', 'z', 'a', 'pdgid']
                }
            })
        return df

    # TODO: list of impacted collimators


    # TODO: does not work when multiple children
    def interactions_per_collimator(self, collimator=0, *, turn=None):
        if isinstance(collimator, str):
            collimator = self._collimator_id(collimator)
        mask = (self._inter > 0) & (self.at_element == collimator)
        if turn is not None:
            mask = mask & (self.at_turn == turn)
            df = pd.DataFrame({
                    'int':  [shortcuts[inter] for inter in self._inter[mask]],
                    'pid':  self.id_before[mask]
                })
            return df.groupby('pid', sort=False, group_keys=False)['int'].agg(list)
        else:
            df = pd.DataFrame({
                    'int':   [shortcuts[inter] for inter in self._inter[mask]],
                    'turn':  self.at_turn[mask],
                    'pid':   self.id_before[mask]
                })
            return df.groupby(['pid', 'turn'], sort=False, group_keys=False)['int'].apply(list)

    def first_touch_per_turn(self, frame=None):
        n_rows = self._index.num_recorded
        df = pd.DataFrame({'id_before': self.id_before[:n_rows],
                           'at_turn': self.at_turn[:n_rows],
                           'at_element': self.at_element[:n_rows]})
        mask = np.char.startswith(self.interaction_type[:n_rows], 'Enter Jaw')
        idx_first = [group.at_element.idxmin() for _, group in df[mask].groupby(
                        ['at_turn', 'id_before'], sort=False, group_keys=False)]
        df_first = self.to_pandas(frame=frame).loc[idx_first]
        df_first.insert(2, "jaw", df_first.interaction_type.astype(str).str[-1])
        to_drop = ['interaction_type',
                   *[col for col in df_first.columns if col.endswith('_after')]]
        to_rename = {col: col.replace('_before', '') for col in df_first.columns if col.endswith('before')}
        to_rename['id_before'] = 'pid'
        return df_first.drop(columns=to_drop).rename(columns=to_rename)


def _get_xcoll_elements(line=None, elements=None, names=None):
    from xcoll.beam_elements import block_classes
    if names is not None and names is not False and \
    (not hasattr(names, '__iter__') or isinstance(names, str)):
        names = [names]
    if elements is not None and elements is not False and \
    (not hasattr(elements, '__iter__') or isinstance(elements, str)):
        elements = [elements]
    if line is None:
        if elements is None:
            raise ValueError("No line nor elements provided!")
    else:
        if elements is not None and elements is not False:
            raise ValueError("Cannot provide both line and elements!")
        if names is None or names is True:
            elements, names = line.get_elements_of_type(block_classes)
            if len(names) == 0:
                raise ValueError("No Xcoll elements in line!")
        elif names is False:
            names = []
            elements = []
        else:
            assert elements is not False
            for name in names:
                if name not in line.element_names:
                    raise ValueError(f"Element {name} not found in line!")
            elements = [line[name] for name in names]
    for idx, element in enumerate(elements):
        if not isinstance(element, block_classes):
            name = name[idx] if names is not None else element.__class__.__name__
            raise ValueError(f"Element {name} not an Xcoll element!")
    return elements, names




# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .general import _pkg_root, __version__, citation

from .beam_elements import (BlackAbsorber, BlackCrystal, TransparentCollimator, TransparentCrystal,
                            EverestBlock, EverestCollimator, EverestCrystal, Geant4Collimator,
                            Geant4CollimatorTip, Geant4Crystal, FlukaCollimator, FlukaCrystal,
                            BlowUp, EmittanceMonitor, collimator_classes, crystal_classes,
                            element_classes)
from .materials import Material, RefMaterial
from .scattering_routines.fluka import FlukaPrototype, FlukaAssembly, create_generic_assembly
from .scattering_routines.geant4 import Geant4Engine
from .colldb import CollimatorDatabase
from .interaction_record import InteractionRecord
from .rf_sweep import RFSweep, prepare_rf_sweep
from .lossmap import LossMap, MultiLossMap
from .particles_tree import ParticlesTree

from .constants import particle_states, particle_state_names, interactions, interaction_names

# Initialise FLUKA environment
from .scattering_routines.fluka.wrapper import FlukaWrapper as _FlukaWrapper
fluka = _FlukaWrapper()

# Initialise Geant4 environment
from .scattering_routines.geant4.wrapper import Geant4Wrapper as _Geant4Wrapper
geant4 = _Geant4Wrapper()

# print("If you use Xcoll in your simulations, please cite us :-)")
# print(citation)



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import xtrack as xt
import xobjects as xo
from .base import BaseCollimator, BaseCrystal, InvalidXcoll
from ..scattering_routines.geometry import XcollGeometry
from ..general import _pkg_root


class TransparentCollimator(BaseCollimator):
    _xofields = { **BaseCollimator._xofields,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = False
    allow_track = True
    allow_double_sided = True
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _noexpr_fields         = BaseCollimator._noexpr_fields
    _skip_in_to_dict       = BaseCollimator._skip_in_to_dict
    _store_in_to_dict      = BaseCollimator._store_in_to_dict
    _internal_record_class = BaseCollimator._internal_record_class

    _depends_on = [BaseCollimator, XcollGeometry]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','transparent_collimator.h')
    ]

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            kwargs.setdefault('_tracking', True)
        super().__init__(**kwargs)
        if not isinstance(self._context, xo.ContextCpu):
            raise ValueError('TransparentCollimator is currently not supported on GPU.')

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context, _buffer=_buffer, _offset=_offset)


class TransparentCrystal(BaseCrystal):
    _xofields = { **BaseCrystal._xofields,
        '_tracking':        xo.Int8
    }

    isthick = True
    needs_rng = False
    allow_track = True
    allow_double_sided = False
    behaves_like_drift = True
    allow_rot_and_shift = False
    allow_loss_refinement = True
    skip_in_loss_location_refinement = True

    _noexpr_fields         = BaseCrystal._noexpr_fields
    _skip_in_to_dict       = BaseCrystal._skip_in_to_dict
    _store_in_to_dict      = BaseCrystal._store_in_to_dict
    _internal_record_class = BaseCrystal._internal_record_class

    _depends_on = [BaseCrystal, XcollGeometry]

    _extra_c_sources = [
        _pkg_root.joinpath('beam_elements','elements_src','transparent_crystal.h')
    ]

    def __init__(self, **kwargs):
        if '_xobject' not in kwargs:
            kwargs.setdefault('_tracking', True)
        super().__init__(**kwargs)
        if not isinstance(self._context, xo.ContextCpu):
            raise ValueError('TransparentCrystal is currently not supported on GPU.')

    def get_backtrack_element(self, _context=None, _buffer=None, _offset=None):
        return InvalidXcoll(length=-self.length, _context=_context, _buffer=_buffer, _offset=_offset)




# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import atexit
import numpy as np

from .prototype import FlukaPrototype, FlukaAssembly
from ...compare import deep_equal


# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
#
# DOES NOT WORK WITH PARALLEL JOBS: Exit handler of one job will delete prototype
# as soon as finished, even if other job still running...
#
# TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
# def exit_handler():
#     """Remove generic assemblies on exit."""
#     for assembly in FlukaPrototype._registry:
#         if isinstance(assembly, FlukaAssembly) \
#         and assembly.fedb_series == 'generic':
#             if len(assembly.elements) == 0:
#                 print(f"Deleting generic FlukaAssembly {assembly.fedb_tag}...")
#                 assembly.delete()
#     # If there are broken assemblies, there might be prototypes left
#     for prototype in FlukaPrototype._registry:
#         if isinstance(prototype, FlukaPrototype) \
#         and prototype.fedb_series == 'generic':
#             if len(prototype.elements) == 0:
#                 print(f"Deleting generic FlukaPrototype {prototype.fedb_tag}...")
#                 prototype.delete()
# atexit.register(exit_handler)


_generic_required_fields = ['material', 'length']
_generic_optional_fields = {'side': 'both', 'width': 0.2, 'height': 0.2}
_generic_crystal_required_fields = ['material', 'length', 'bending_radius']
_generic_crystal_optional_fields = {'side': 'left', 'width': 0.02, 'height': 0.05}


def create_generic_assembly(**kwargs):
    _validate_kwargs(kwargs)
    # Check if the assembly already exists
    for prototype in FlukaPrototype._registry:
        if not isinstance(prototype, FlukaAssembly):
            continue
        if prototype.fedb_series == 'generic':
            found = True
            for field in _generic_required_fields + list(_generic_optional_fields.keys()):
                if not(deep_equal(kwargs[field], getattr(prototype, field))):
                    found = False
                    break
            if kwargs['is_crystal']:
                for field in _generic_crystal_required_fields + list(_generic_crystal_optional_fields.keys()):
                    if not(deep_equal(kwargs[field], getattr(prototype, field))):
                        found = False
                        break
            if found and prototype.exists():
                return prototype
    # Get an ID
    existing_ids = [int(p.fedb_tag[3:]) for p in FlukaPrototype._registry
                    if p.fedb_series == 'generic' and isinstance(p, FlukaAssembly)]
    max_id = max(existing_ids) if existing_ids else 0
    new_id = min(set(range(max_id+2)) - set(existing_ids))
    fedb_series = 'generic'
    fedb_tag = f'GEN{new_id:03d}'
    # Create the objects
    tank = FlukaPrototype(fedb_series, f'{fedb_tag}_T', _allow_generic=True)
    kwargs_body = kwargs.copy()
    kwargs_body.pop('side')
    body = FlukaPrototype(fedb_series, f'{fedb_tag}_B', _allow_generic=True, **kwargs_body)
    assm = FlukaAssembly(fedb_series, fedb_tag, _allow_generic=True, **kwargs)
    assm._prototypes = [tank, body]
    return assm


def _validate_kwargs(kwargs):
    kwargs.setdefault('is_crystal', False)
    from ...materials import _resolve_material
    kwargs['material'] = _resolve_material(kwargs['material'], ref='fluka', allow_none=False)
    if kwargs['is_crystal']:
        for field in _generic_crystal_required_fields:
            if field not in kwargs or kwargs[field] is None:
                raise ValueError(f"Need to provide {field}!")
        for field, opt_value in _generic_crystal_optional_fields.items():
            if field not in kwargs or kwargs[field] is None:
                kwargs[field] = opt_value
    else:
        for field in _generic_required_fields:
            if field not in kwargs or kwargs[field] is None:
                raise ValueError(f"Need to provide {field}!")
        for field, opt_value in _generic_optional_fields.items():
            if field not in kwargs or kwargs[field] is None:
                kwargs[field] = opt_value
    if kwargs.get('side') not in ['both', 'left', 'right']:
        raise ValueError(f"Side must be 'both', 'left' or 'right', but got {kwargs.get('side')}!")
    if kwargs['width'] > 0.25:
        kwargs['width'] = 0.25
    if kwargs['height'] > 0.25:
        kwargs['height'] = 0.25
    kwargs.pop('fedb_series', None) # fedb_series is always 'generic'
    kwargs.pop('fedb_tag', None)    # fedb_tag is always 'GENnnn'
    kwargs['angle'] = 0             # Only horizontal assembly (angle will be provided by LineBuilder)


def _assembly_file(fedb, fedb_tag, side, **kwargs):
    template_assembly = f"""\
# --------------------------------------------------------------------------------------------------------------
PROTOTYPE       {fedb_tag}_T
# --------------------------------------------------------------------------------------------------------------
# prototype of the tank:
FEDB_TAG        {fedb_tag}_T                # tag in the fedb (for the name of the file)
FEDB_SERIES     generic                # series of the fedb (for the name of the file)
# this rot-defi is only for local use:
ROT-DEFI         0.0       0.0       0.0     100.0   -3000.0    1000.0 proto
#
# --------------------------------------------------------------------------------------------------------------
PROTOTYPE       {fedb_tag}_B
# --------------------------------------------------------------------------------------------------------------
# prototype of the jaw:
FEDB_TAG        {fedb_tag}_B                # tag in the fedb (for the name of the file)
FEDB_SERIES     generic                # series of the fedb (for the name of the file)
# this rot-defi is only for local use:
ROT-DEFI         0.0       0.0       0.0       0.0   -3000.0    1000.0 proto
#
#
#
# --------------------------------------------------------------------------------------------------------------
ASSEMBLY        {fedb_tag}
# --------------------------------------------------------------------------------------------------------------
#
# needed bodies:
#           rename      rotname     name_in_file fedb_series fedb_tag    index
BODY        CONTAINO    CONTAINO    {fedb_tag}_T     generic     {fedb_tag}_T    1
BODY        CONTAINI    CONTAINO    {fedb_tag}_I     generic     {fedb_tag}_T    1"""
    if side in [None, 'both', 'left']:
        template_assembly += f"""
* jaw on positive x:
BODY        JAW_POS     JAW_POS     {fedb_tag}_B     generic     {fedb_tag}_B    1"""
    if side in [None, 'both', 'right']:
        template_assembly += f"""
* jaw on negative x:
BODY        JAW_NEG     JAW_NEG     {fedb_tag}_B     generic     {fedb_tag}_B    1"""
    template_assembly += f"""
#
# define regions:
#           rename      material    rotbody     defition
REGION      *           EXTERNAL    *           -CONTAINO
* tank:
REGION      TANK        LATTICE     CONTAINO    +CONTAINO -CONTAINI
* vacuum between jaws:"""
    subtract_jaws = ''
    if side in [None, 'both', 'left']:
        subtract_jaws += ' -JAW_POS'
    if side in [None, 'both', 'right']:
        subtract_jaws += ' -JAW_NEG'
    template_assembly += f"""
REGION      INNERVAC    VACUUM      *           +CONTAINI{subtract_jaws}"""
    if side in [None, 'both', 'left']:
        template_assembly += f"""
* jaw on positive x:
REGION      JAW_POS     LATTICE     JAW_POS     +JAW_POS"""
    if side in [None, 'both', 'right']:
        template_assembly += f"""
* jaw on negative x:
REGION      JAW_NEG     LATTICE     JAW_NEG     +JAW_NEG"""
    template_assembly += f"""
#
ROT-DEFI             0.0         0.0         0.0         0.0         0.0         0.0 CONTAINO"""
    if side in [None, 'both', 'left']:
        template_assembly += f"""
ROT-DEFI             0.0         0.0         0.0         0.0         0.0         0.0 JAW_POS"""
    if side in [None, 'both', 'right']:
        template_assembly += f"""
* rotate by 180 deg the negative jaw:
ROT-DEFI           300.0         0.0       180.0         0.0         0.0         0.0 JAW_NEG
"""
    return _write_file(fedb, "assemblies", f"generic_{fedb_tag}.lbp",
                       template_assembly)


def _body_file(fedb, fedb_tag, length, width, height, **kwargs):
    template_body = f"""\
RPP {fedb_tag}_B   0.0 {100*width} -{100*height/2} {100*height/2} -{length*100/2} {length*100/2}
"""
    body_file = _write_file(fedb, "bodies", f"generic_{fedb_tag}_B.bodies",
                            template_body)

    # Tank body should fit in blackhole (0.8m x 0.8m) for any angle, so maximally 0.8*sqrt(2)/2 = 0.565 for each side
    template_tank = f"""\
RPP {fedb_tag}_T  -28 28 -28 28 -{length*100 + 5} {length*100 + 5}
RPP {fedb_tag}_I  -28 28 -28 28 -{length*100 + 5} {length*100 + 5}
* For BB, need less margin!
*RPP {fedb_tag}_T  -28 28 -28 28 -{length*100 + 1e-12} {length*100 + 1e-12}
*RPP {fedb_tag}_I  -28 28 -28 28 -{length*100 + 1e-12} {length*100 + 1e-12}
"""
    tank_file = _write_file(fedb, "bodies", f"generic_{fedb_tag}_T.bodies",
                            template_tank)
    return body_file, tank_file


def _region_file(fedb, fedb_tag, **kwargs):
    template_body_reg = f"""\
{fedb_tag}_B     5 +{fedb_tag}_B
"""
    body_file = _write_file(fedb, "regions", f"generic_{fedb_tag}_B.regions",
                            template_body_reg)

    template_tank_reg = f"""\
{fedb_tag}_T     5 +{fedb_tag}_T -{fedb_tag}_I
{fedb_tag}_I     5 +{fedb_tag}_I
"""
    tank_file = _write_file(fedb, "regions", f"generic_{fedb_tag}_T.regions",
                            template_tank_reg)
    return body_file, tank_file


def _material_file(fedb, fedb_tag, material, **kwargs):
    mat = material.fluka_name
    template_body_mat = f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
ASSIGNMA    {mat:>8}  {fedb_tag:>6}_B
"""
    body_file = _write_file(fedb, "materials", f"generic_{fedb_tag}_B.assignmat",
                            template_body_mat)

    template_tank_mat = f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
ASSIGNMA      VACUUM  {fedb_tag:>6}_T
ASSIGNMA      VACUUM  {fedb_tag:>6}_I
"""
    tank_file = _write_file(fedb, "materials", f"generic_{fedb_tag}_T.assignmat",
                            template_tank_mat)
    return body_file, tank_file

def _crystal_body_file(fedb, fedb_tag, length, bending_radius, width, height, **kwargs):
    template_body = f"""\
RPP {fedb_tag}_B   0.0 {width*(100+10)} -{height*(100+10)/2} {height*(100+10)/2} -{length*(100+20)} {length*(100+20)}
YCC {fedb_tag}Z1  0.0 {bending_radius*100} {bending_radius*100}
YCC {fedb_tag}Z2  0.0 {bending_radius*100} {bending_radius*100-width*100}
PLA {fedb_tag}P1  1.0 0.0 {np.cos(length/bending_radius)/np.sin(length/bending_radius)} {bending_radius*100} 0.0 0.0
XYP {fedb_tag}P2  0.0
"""
    body_file = _write_file(fedb, "bodies", f"generic_{fedb_tag}_B.bodies",
                            template_body)
    template_tank = f"""\
RPP {fedb_tag}_T  -28 28 -28 28 -{length*100 + 5} {length*100 + 5}
RPP {fedb_tag}_I  -28 28 -28 28 -{length*100 + 5} {length*100 + 5}
"""
    tank_file = _write_file(fedb, "bodies", f"generic_{fedb_tag}_T.bodies",
                            template_tank)
    return body_file, tank_file


def _crystal_region_file(fedb, fedb_tag, **kwargs):
    template_body_reg = f"""\
{fedb_tag}_B     5 | +{fedb_tag}_B +{fedb_tag}Z1 -{fedb_tag}Z2 +{fedb_tag}P1 - {fedb_tag}P2
{fedb_tag}B2     5 | +{fedb_tag}_B +{fedb_tag}Z2
                   | +{fedb_tag}_B -{fedb_tag}Z1
                   | +{fedb_tag}_B -{fedb_tag}P1
                   | +{fedb_tag}_B +{fedb_tag}P2 -{fedb_tag}Z2
"""
    body_file = _write_file(fedb, "regions", f"generic_{fedb_tag}_B.regions",
                            template_body_reg)
    template_body_tank = f"""\
{fedb_tag}_T     5 | +{fedb_tag}_T -{fedb_tag}_I
{fedb_tag}_I     5 | +{fedb_tag}_I
"""
    tank_file = _write_file(fedb, "regions", f"generic_{fedb_tag}_T.regions",
                            template_body_tank)
    return body_file, tank_file


def _crystal_material_file(fedb, fedb_tag, material, **kwargs):
    mat = material.fluka_name
    template_body_mat = f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
ASSIGNMA    {mat:>8}  {fedb_tag:>6}_B
ASSIGNMA      VACUUM  {fedb_tag:>6}B2
"""
    body_file = _write_file(fedb, "materials", f"generic_{fedb_tag}_B.assignmat",
                            template_body_mat)
    template_tank_mat = f"""\
* ..+....1....+....2....+....3....+....4....+....5....+....6....+....7..
ASSIGNMA      VACUUM  {fedb_tag:>6}_T
ASSIGNMA      VACUUM  {fedb_tag:>6}_I
"""
    tank_file = _write_file(fedb, "materials", f"generic_{fedb_tag}_T.assignmat",
                            template_tank_mat)
    return body_file, tank_file


def _write_file(fedb, directory, filename, content):
    path = fedb / directory / filename
    with path.open('w') as fid:
        fid.write(content)
    return path



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from __future__ import annotations

import numpy as np
from typing import Iterable, Callable, Optional, Dict, List, Tuple

from xtrack.particles import LAST_INVALID_STATE


class ParticlesTree:
    def __init__(self, particle_ids, parent_ids=None, *, mask=None):
        """
        Usage:
          - ParticlesTree(particle_ids, parent_ids)
          - ParticlesTree(part)                         # uses all particles
          - ParticlesTree(part, mask=...)               # custom boolean mask
        """

        print("WARNING: ParticlesTree is an experimental feature and the API "
              "will most probably change in future releases.")

        # ---- Accept either arrays or an xtrack.Particles-like object ----
        if parent_ids is None and hasattr(particle_ids, "particle_id") and hasattr(particle_ids, "parent_particle_id"):
            part = particle_ids
            if mask is None:
                if hasattr(part, "state"):
                    mask = np.asarray(part.state) > LAST_INVALID_STATE
                else:
                    mask = slice(None)  # use all rows
            ids = np.asarray(part.particle_id)[mask]
            par = np.asarray(part.parent_particle_id)[mask]
        else:
            ids = np.asarray(particle_ids)
            par = np.asarray(parent_ids)
            if par is None:
                raise ValueError("When passing arrays, both particle_ids and parent_ids are required.")

        # ---- Basic validation / normalisation ----
        if ids.ndim != 1 or par.ndim != 1 or ids.shape != par.shape:
            raise ValueError("ids and parent_ids must be 1D arrays of equal length")
        ids = ids.astype(np.int64, copy=False)
        par = par.astype(np.int64, copy=False)
        n = ids.size

        if n == 0:
            self.ids = ids
            self.parent_idx = np.zeros(0, dtype=np.int64)
            self.root_idx = np.zeros(0, dtype=np.int64)
            self.root_ids = np.zeros(0, dtype=np.int64)
            self.children_ptr = np.zeros(1, dtype=np.int64)
            self.children_idx = np.zeros(0, dtype=np.int64)
            self._order = None
            self._ids_sorted = None
            return

        # ---- id -> index (one sort) ----
        order = np.argsort(ids)
        ids_sorted = ids[order]
        parent_idx = order[np.searchsorted(ids_sorted, par)]

        # ---- pointer jumping to roots ----
        P = parent_idx.copy()
        max_iters = int(np.ceil(np.log2(max(1, n)))) + 3
        for _ in range(max_iters):
            newP = P[P]
            if np.array_equal(newP, P):
                break
            P = newP
        else:
            raise ValueError("Parent chains did not converge (cycle suspected).")

        root_idx = P
        root_ids = ids[root_idx]

        # ---- children CSR (exclude self-loops) ----
        self_idx = np.arange(n, dtype=np.int64)
        mask_child = parent_idx != self_idx
        if np.any(mask_child):
            parents = parent_idx[mask_child]
            children = np.nonzero(mask_child)[0]
            k = np.argsort(parents)
            parents = parents[k]
            children = children[k]

            counts = np.zeros(n, dtype=np.int64)
            uniq, cnts = np.unique(parents, return_counts=True)
            counts[uniq] = cnts

            ptr = np.empty(n + 1, dtype=np.int64)
            np.cumsum(counts, out=ptr[1:])
            ptr[0] = 0
            csr_idx = children
        else:
            ptr = np.zeros(n + 1, dtype=np.int64)
            csr_idx = np.zeros(0, dtype=np.int64)

        # ---- store ----
        self.ids = ids
        self.parent_idx = parent_idx
        self.root_idx = root_idx
        self.root_ids = root_ids
        self.children_ptr = ptr
        self.children_idx = csr_idx

        # lazy lookup cache (always 1-D)
        self._order = None
        self._ids_sorted = None

    # keep the rest of the class as before …
    def index_of(self, particle_id: int) -> int:
        if self._order is None:
            self._order = np.argsort(self.ids)
            self._ids_sorted = self.ids[self._order].ravel()  # ensure 1-D
        j = int(np.searchsorted(self._ids_sorted, int(particle_id)))
        if j >= self._ids_sorted.size or self._ids_sorted[j] != particle_id:
            raise KeyError(f"particle_id {particle_id} not found")
        return int(self._order[j])

    def indices_of(self, particle_ids: Iterable[int]) -> np.ndarray:
        x = np.asarray(list(particle_ids))
        if self._order is None:
            self._order = np.argsort(self.ids)
            self._ids_sorted = self.ids[self._order]
        j = np.searchsorted(self._ids_sorted, x)
        ok = (j < self._ids_sorted.size) & (self._ids_sorted[j] == x)
        if not np.all(ok):
            missing = x[~ok]
            raise KeyError(f"particle_id(s) not found: {missing.tolist()}")
        return self._order[j]

    def children_of_index(self, i: int) -> np.ndarray:
        a, b = int(self.children_ptr[i]), int(self.children_ptr[i + 1])
        return self.children_idx[a:b]

    def descendants_indices(self, i: int) -> np.ndarray:
        out = []
        stack = [int(i)]
        while stack:
            u = stack.pop()
            ks = self.children_of_index(u)
            if ks.size:
                out.extend(ks.tolist())
                stack.extend(ks.tolist())
        return np.asarray(out, dtype=np.int64)

    def descendants_ids(self, particle_id: int) -> np.ndarray:
        i = self.index_of(particle_id)
        return self.ids[self.descendants_indices(i)]

    def chain_of_id(self, particle_id: int, include_self: bool = True) -> List[int]:
        i = self.index_of(particle_id)
        chain: List[int] = []
        if include_self:
            chain.append(int(self.ids[i]))
        while True:
            p = int(self.parent_idx[i])
            if p == i:
                break
            chain.append(int(self.ids[p]))
            i = p
        return chain

    # ---------- ASCII tree (sorted) ----------
    def print_tree(self,
                   root_particle_id: Optional[int] = None,
                   label: Optional[Callable[[int], str]] = None,
                   max_depth: Optional[int] = None,
                   sort_children: bool = True,
                   sort_roots: bool = True) -> None:
        """
        Print an ASCII tree/forest. Children and roots can be sorted by particle_id.
        """
        if label is None:
            label = lambda pid: str(pid)

        def kids(i: int) -> np.ndarray:
            k = self.children_of_index(i)
            if sort_children and k.size:
                k = k[np.argsort(self.ids[k])]
            return k

        def emit(i: int, prefix: str, depth: int, is_last: bool):
            connector = "" if depth == 0 else ("└─ " if is_last else "├─ ")
            print(prefix + connector + label(int(self.ids[i])))
            if max_depth is not None and depth >= max_depth:
                return
            ks = kids(i)
            if ks.size == 0:
                return
            new_prefix = prefix + ("" if depth == 0 else ("   " if is_last else "│  "))
            for j, c in enumerate(ks):
                emit(int(c), new_prefix, depth + 1, j == ks.size - 1)

        if root_particle_id is None:
            roots = self.originals_indices()
            if sort_roots and roots.size:
                roots = roots[np.argsort(self.ids[roots])]
            for idx_r, r in enumerate(roots):
                emit(int(r), "", 0, True)
                if idx_r != len(roots) - 1:
                    print()  # blank line between trees
        else:
            emit(self.index_of(root_particle_id), "", 0, True)

    # ---------- layout (pure NumPy; sorted) ----------
    def layout_positions(self,
                         root_particle_id: Optional[int] = None,
                         max_depth: Optional[int] = None,
                         sort_children: bool = True,
                         sort_roots: bool = True,
                         x_gap: float = 1.0,
                         y_gap: float = 1.0) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Compute a tidy top-down layout. Returns (node_idx, x, y, depth) for the drawn nodes.
        """
        n = self.ids.size
        # roots
        if root_particle_id is None:
            roots = self.originals_indices()
            if sort_roots and roots.size:
                roots = roots[np.argsort(self.ids[roots])]
            roots = roots.tolist()
        else:
            roots = [self.index_of(int(root_particle_id))]

        def _kids(i: int) -> np.ndarray:
            k = self.children_of_index(i)
            if sort_children and k.size:
                k = k[np.argsort(self.ids[k])]
            return k

        depth = np.zeros(n, dtype=np.int64)
        seen = np.zeros(n, dtype=bool)
        post: List[int] = []

        # DFS to set depths and produce postorder
        stack: List[Tuple[int, int, int]] = []
        for r in roots:
            stack.append((int(r), 0, 0))
            while stack:
                u, st, d = stack.pop()
                if st == 0:
                    if seen[u]:
                        continue
                    seen[u] = True
                    depth[u] = d
                    stack.append((u, 1, d))
                    ks = _kids(u)
                    if max_depth is not None and d >= max_depth:
                        ks = ks[:0]
                    for c in ks[::-1]:
                        stack.append((int(c), 0, d + 1))
                else:
                    post.append(u)

        # (leaf widths computed implicitly by placing children first)
        x = np.zeros(n, dtype=float)
        y = depth.astype(float) * y_gap
        ordered_nodes: List[int] = []
        cur_x = 0.0

        def place(u: int):
            nonlocal cur_x
            ks = _kids(u)
            if max_depth is not None and depth[u] >= max_depth:
                ks = ks[:0]
            if ks.size == 0:
                x[u] = cur_x
                cur_x += x_gap
            else:
                left_edge = cur_x
                for c in ks:
                    place(int(c))
                right_edge = cur_x - x_gap
                x[u] = (left_edge + right_edge) / 2.0
            ordered_nodes.append(u)

        for r in roots:
            place(int(r))
            cur_x += x_gap  # gap between trees

        node_idx = np.array(ordered_nodes, dtype=np.int64)
        return node_idx, x[node_idx], y[node_idx], depth[node_idx]

    # ---------- plotting (Matplotlib only; sorted) ----------
    def plot_tree(self,
                  root_particle_id: Optional[int] = None,
                  max_depth: Optional[int] = None,
                  sort_children: bool = True,
                  sort_roots: bool = True,
                  with_labels: bool = True,
                  label: Optional[Callable[[int], str]] = None,
                  node_marker: str = "o",
                  node_size: float = 50.0,
                  linewidth: float = 1.0,
                  ax=None):
        """
        Draw a subtree or whole forest using Matplotlib.
        """
        import matplotlib.pyplot as plt

        if label is None:
            label = lambda pid: str(pid)

        idx, xs, ys, depth = self.layout_positions(root_particle_id=root_particle_id,
                                                   max_depth=max_depth,
                                                   sort_children=sort_children,
                                                   sort_roots=sort_roots)
        slot = {int(i): k for k, i in enumerate(idx)}

        # edges
        for i in idx:
            p = int(self.parent_idx[i])
            if p == i:
                continue
            if int(i) not in slot or int(p) not in slot:
                continue
            x0, y0 = xs[slot[p]], ys[slot[p]]
            x1, y1 = xs[slot[i]], ys[slot[i]]
            ax = ax or plt.gca()
            ax.plot([x0, x1], [y0, y1], linewidth=linewidth)

        # nodes
        ax = ax or plt.gca()
        ax.scatter(xs, ys, s=node_size, marker=node_marker)

        if with_labels:
            for i, x0, y0 in zip(idx, xs, ys):
                ax.text(x0, y0, label(int(self.ids[int(i)])), ha="center", va="bottom")

        ax.invert_yaxis()
        ax.set_aspect("equal", adjustable="box")
        ax.axis("off")
        return ax



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .material import Material, RefMaterial, _DEFAULT_MATERIAL, _resolve_material
from .atoms import *
from .allotropes import *
from .compounds import *
from .mixtures import *
from .sixtrack import *

# Freeze all material instances (to prevent accidental modification)
for name, obj in list(globals().items()):  # Have to wrap in list to take a snapshot (avoid updating in-place)
    if isinstance(obj, Material) and name != 'obj':
        obj._frozen = True
del name, obj

# Get database instance and expose some methods
from .database import db
show = db.show

# Some sources:
# https://epaper.kek.jp/ipac2023/pdf/WEPA148.pdf
# https://pdf.sciencedirectassets.com/271508/1-s2.0-S0008622318X00074/1-s2.0-S0008622318303555/main.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEJD%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FwEaCXVzLWVhc3QtMSJGMEQCIEpatVR2iMAynQtBCSVxWwxx0WFnuPyFK%2B3jgOJ2INnhAiBBw6VGFuLSEoweOC7ryGAImpGk8ApFmp%2BBYwt3hQFXEyqyBQg5EAUaDDA1OTAwMzU0Njg2NSIMcqyr5hKtD3HLV7%2F2Ko8FQOHztlojw87Oi7LV2xR%2Bkwk9JkxSumwlU4aqcgWrDL%2BHnheO7yujorf9v3HMoJE906w4h8ph2cj2KbX7USxL7x0aHkDG8Sm0%2Fx7lRohiq3oQtWF4WucsVEvai%2BiyKff%2FTdUNFhaR%2BICPjMA4Tl%2F5YepW6lgDP1S7UBwCkxl4ymSdbCS8Hbtevx%2FyJIYoEuZHinZfv6yCX0w1iqlvCCaUb2m9haj%2BCHQjCBtqClnQF463e%2FC4j8EApMoidsggxBGXEqNlc%2FuxAcJz4YJuv5a5dK4L9g3mOhBaUFLf5tmSmMptQG9BlIr8MdVCsrOdjbn5u0ho9g%2BWCzEap0YR%2FUKmZ7PkcmZalTPTglyBy%2BouXmiJJ1QSU7mPw%2BXEizno8572zjv8eA6a7aziCwG3tHjKXabe%2FQrHu%2BzoxkF5TgpzyPPuWLxf5XAGrRMBrXju4b8u1eYsQQTAelTBlxtT9WSg9J0M1fFAL%2Bq870p0OtZsTwMWqN6025%2FD6FbwmItW%2FTS4GIBS%2B2E0nRIQvZn2mq3HL3w2okPFjEZVluR1q3vH5NTUB9KSFeQROFjVUoC4hDHKE3fgxX8hfPboKFbN73MrkMI65x0vj42PrqmbQvWHxV16ZtOrBVMKfiIW5ffK6te9Zyi%2FCC7%2BFQKwOfZg6J7O3pSIN5tKQDmXwV%2FuxAvYEK2iLJzh3s6oMSOZdYWeKoy1of3D0TzvajHCThM99hLWt19YEJzj%2B2EE9jUfbRv1n1Ctu%2FKEQpYYYV2qF7Yvw7ivwscTOrDikRjnY%2F2SE6C6kOynmgoXoJ3IiDxN5a%2BZto3B4y6ZbRPacmm96Pr5jJSDksq9n9Gb56kRg%2FyPoACGUkhSsVra8aISegF87wPMaDC897DHBjqyAUxY1XGbgOrxrC2NO1l7QQE0tMrmXr23nsZJ0htjCkO7FmG8Q2PreWfasjJuK0EpwwnDEMdzdvG4GSJMngSBLrFrXs7R%2F1%2Fj8gSd4ZuMjkdY1E22qORRSi9MSaDi1mYgEu6W5S9Kq5u5avLz2ZvrT5pQn0rd2JhZpQpl9lyNp67CYGbX5PzYIpKJ5Yw8msrTH5wxt8tagssiB6xHnpgnsDLo%2BDxXVXlwd5FLI9KNB1MlZro%3D&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20251012T235435Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAQ3PHCVTY7NQSSLCB%2F20251012%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=378e6b9cacb299d5ebc48df11583b1db9a56a9c38ea4eb0088fa25181d79bc5e&hash=ba8572e278d3740d984f493cdcc196226e09d391f2bfbd245654826268e741d3&host=68042c943591013ac2b2430a89b270f6af2c76d8dfd086a07176afe7c76c2c61&pii=S0008622318303555&tid=spdf-dc8d6dd3-57b0-41ef-b769-ee6273970373&sid=69e545f78f7a2947339ac3f-be9cff1d6e6fgxrqb&type=client&tsoh=d3d3LnNjaWVuY2VkaXJlY3QuY29t&rh=d3d3LnNjaWVuY2VkaXJlY3QuY29t&ua=05155858020e5e5003&rr=98da85f3cd01bc81&cc=fr
# https://indico.cern.ch/event/1076808/contributions/5047893/attachments/2511128/4316133/Jaw-materials-justifications_v5_NM.xlsx
# https://edms.cern.ch/ui/#!master/navigator/document?D:101081798:101081798:approvalAndComments
# https://iopscience.iop.org/article/10.1088/1742-6596/2687/8/082037/pdf
# https://indico.cern.ch/event/340703/contributions/802090/attachments/668678/919103/2015_26_03_collmat.pdf


# TCP: CarbonFibreCarbon AC150GPH
# TCPPM: MolybdenumGraphite MG6403Fc
# TCSG: AC150GPH
# TCSP: AC150GPH
# TCSPM: MG6403Fc
# TCT: INERM180
# TCL: COPPER
# TCLA: INERM180
# TDI: BN5000 / Al / Cu
# TCLIA: GraphiteR4550 GRAR4550
# TCLIB: AC150GPH
# TCDQ: CC_1_75 / CC_1_40 / CC_1_75



# copyright ############################### #
# This file is part of the Xcoll Package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

import numpy as np

import xtrack as xt
import xtrack.particles.pdg as pdg

from ...constants import (LOST_ON_FLUKA_COLL, MASSLESS_OR_NEUTRAL,
                          VIRTUAL_ENERGY, HIT_ON_FLUKA_COLL)


def track_pre(coll, particles):
    import xcoll as xc

    # Initialize ionisation loss accumulation variable
    if coll._acc_ionisation_loss < 0:
        coll._acc_ionisation_loss = 0.

    if not xc.fluka.engine.assert_ready_to_track_or_skip(
    coll, particles, _necessary_attributes=['fluka_id']):
        return False  # Stop tracking

    if not xc.fluka.engine._flukaio_connected:
        raise ValueError(f"FlukaEngine not yet running!\nPlease do this first, by calling "
                       + f"xcoll.FlukaEngine.start(fluka_input_file.inp). "
                       + f"(id: {id(xc.fluka.engine)})")

    npart = particles._num_active_particles
    if 1.4*npart > xc.fluka.engine.capacity:
        raise ValueError(f"Tracking {npart} particles but only {xc.fluka.engine.capacity} allocated in "
                       + f"FlukaEngine!\nRemember to leave room for secondaries...")
    if xc.fluka.engine.relative_capacity == 2 and xc.fluka.engine.particle_ref.pdg_id[0] != 2212:
        xc.fluka.engine._print("Warning: relative_capacity is set to 2. This is "
                             + "probably not enough for anything except protons.")

    xc.fluka.engine.init_tracking(npart+particles._num_lost_particles)

    if particles.particle_id.max() > xc.fluka.engine.max_particle_id:
        raise ValueError(f"Some particles have an id ({particles.particle_id.max()}) "
                       + f"that is higher than the highest id known to FLUKA ("
                       + f"{xc.fluka.engine.max_particle_id}).\nThis could happen if this "
                       + "particles object is larger than the first particles instance "
                       + "tracked in this session, or if secondary particles are generated "
                       + "somewhere else than FLUKA.\nIn that case, call "
                       + "xcoll.FlukaEngine.init_tracking(max_particle_id) before tracking "
                       + "with a value large enough to accommodate secondaries outside of "
                       + "FLUKA.\nIn any case, please stop and restart the FlukaEngine now.")

    coll._drift(particles, -coll.length_front)
    return True  # Continue tracking


def track_post(coll, particles):
    coll._drift(particles, -coll.length_back)
    alive_states = np.unique(particles.state[particles.state > 0])
    assert len(alive_states) <= 1, f"Unexpected alive particle states after tracking: {alive_states}"
    if len(alive_states) == 1:
        assert alive_states[0] == 1, f"Unexpected alive particle state after tracking: {alive_states[0]}"


def _expand(arr, available_capacity, dtype=float):
    return np.concatenate((arr, np.zeros(available_capacity, dtype=dtype)))


def track_core(coll, part):
    import xcoll as xc
    try:
        from pyflukaf import track_fluka
    except (ModuleNotFoundError, ImportError) as error:
        xc.fluka.engine._warn_pyfluka(error)
        return

    send_to_fluka  = part.state == HIT_ON_FLUKA_COLL
    npart          = send_to_fluka.sum()
    max_id         = part.particle_id[part.state > -9999].max()
    assert npart  <= part._num_active_particles
    if npart == 0:
        return

    # Get particle data
    m0         = part.mass0
    q0         = part.q0
    p0c        = part.p0c[0]
    E0         = part.energy0[0]
    beta0      = part.beta0[0]
    gamma0     = part.gamma0[0]
    mass       = part.mass[send_to_fluka]
    charge     = part.charge[send_to_fluka]
    pdg_id     = part.pdg_id[send_to_fluka]
    _, A, Z, _ = pdg.get_properties_from_pdg_id(pdg_id)
    A          = np.array([0 if pdgid < 0 else aa for aa, pdgid in zip(A, pdg_id)])  # FLUKA treats antiprotons as A = 0 = Z
    Z          = np.array([0 if pdgid < 0 else zz for zz, pdgid in zip(Z, pdg_id)])
    precision  = p0c * 1.e-12  # To avoid numerical issues like negative enervy. Ideally this should be 2.22e-15

    # Decide how much extra capacity to send to FLUKA
    min_capacity = 50
    available_capacity = xc.fluka.engine.capacity - max_id - 1
    if available_capacity < min_capacity:
        raise ValueError("Not enough capacity in particles to accomodate secondaries.")
    available_capacity = min(int(np.ceil((xc.fluka.engine.relative_capacity - 1)*npart)),
                         available_capacity)
    available_capacity = max(available_capacity, min_capacity) # Some minimum value to be safe

    # Prepare arrays for FORTRAN
    data = {}
    data['x']      = _expand(part.x[send_to_fluka] * 1000., available_capacity)
    data['xp']     = _expand(part.px[send_to_fluka] * part.rpp[send_to_fluka] * 1000.,
                             available_capacity)
    data['y']      = _expand(part.y[send_to_fluka] * 1000., available_capacity)
    data['yp']     = _expand(part.py[send_to_fluka] * part.rpp[send_to_fluka] * 1000.,
                             available_capacity)
    data['zeta']   = _expand(part.zeta[send_to_fluka] * 1000., available_capacity)
    data['e']      = _expand(part.energy[send_to_fluka] / 1.e6, available_capacity)
    data['m']      = _expand(mass / 1.e6, available_capacity)
    data['q']      = _expand(charge.astype(np.int16), available_capacity, dtype=np.int16)
    data['A']      = _expand(A.astype(np.int32), available_capacity, dtype=np.int32)
    data['Z']      = _expand(Z.astype(np.int32), available_capacity, dtype=np.int32)
    data['pdg_id'] = _expand(pdg_id.astype(np.int32), available_capacity, dtype=np.int32)
    # FLUKA is 1-indexed
    data['pid']    = _expand(part.particle_id[send_to_fluka].astype(np.int32) + 1,
                             available_capacity, dtype=np.int32)
    # FLUKA does not use a parent ID, but a primary ID (hence not the direct parent but the first impact)
    # After one passage, there is no difference between parent ID and primary ID, but when a child gets
    # children in a second passage, we cannot trace them to the correct parent (only to the correct grand-
    # parent). To accommodate this, we do not send the parent ID to FLUKA but keep it in the particles
    # object.
    data['ppid']   = data['pid'].copy()
    data['weight'] = _expand(part.weight[send_to_fluka], available_capacity)
    # TODO: Hard-coded spin (currently not used)
    data['spin_x'] = _expand(np.zeros(npart), available_capacity)
    data['spin_y'] = _expand(np.zeros(npart), available_capacity)
    data['spin_z'] = _expand(np.zeros(npart), available_capacity)

    # Change npart to np.array to make it writable, store some initial data
    npart    = np.array(npart, dtype=np.int64)
    s_in     = part.s[send_to_fluka][0]
    ele_in   = part.at_element[send_to_fluka][0]
    turn_in  = part.at_turn[send_to_fluka][0]
    start    = part.start_tracking_at_element  # TODO: is this needed?

    # send to fluka
    track_fluka(turn=turn_in+1,     # Turn indexing start from 1 with FLUKA IO (start from 0 with xpart)
                fluka_id=coll.fluka_id,
                length=coll.length + coll.length_front + coll.length_back,
                alive_part=npart,
                max_part=npart + available_capacity,
                x_part=data['x'],
                xp_part=data['xp'], # FLUKA uses director cosine. This is exactly equal to px / (1+delta)
                y_part=data['y'],
                yp_part=data['yp'],
                zeta_part=data['zeta'],
                e_part=data['e'],   # FlukaIO pretends this is momentum, but the flukaserver source proves it is energy
                m_part=data['m'],
                q_part=data['q'],
                A_part=data['A'],
                Z_part=data['Z'],
                pdg_id_part=data['pdg_id'],
                part_id=data['pid'],
                parent_id=data['ppid'],
                part_weight=data['weight'],
                spin_x_part=data['spin_x'],
                spin_y_part=data['spin_y'],
                spin_z_part=data['spin_z']
    )

    # Careful with all the masking!
    # Double-mask assignment does not work, e.g. part.state[mask1][mask2] = 1 will do nothing...

    new_pid  = data['pid'][:npart] - 1   # return to python 0-index
    new_ppid = data['ppid'][:npart] - 1  # return to python 0-index

    # TODO: Impact Table
    #     Absorbed: trivial
    #     Not hit:    pid+ppid did not change and dE = 0
    #     Hit (MCS):  pid+ppid did not change and dE != 0
    #     Hit (nucl): above mask + children

    # TODO: FLUKA returns particles that have undergone a nuclear interaction as new particles: we probably want to correct (after registering interaction)

    E_diff = np.zeros(len(part.x))


    # Update existing particles  (these missed the collimator or only underwent elastic interactions)
    # ===============================================================================================
    mask_alive = new_pid <= max_id
    idx_old = np.argsort(part.particle_id)
    pid_old_sorted = part.particle_id[idx_old]
    pos = np.searchsorted(pid_old_sorted, new_pid[mask_alive])
    idx_alive = idx_old[pos]

    # Kill particles that died just now
    # We cannot do part.state[~idx_alive] = LOST_ON_FLUKA_COLL (as idx_alive is not a mask)
    # So instead, we flag all as dead, and correct the alive ones later
    part.state[send_to_fluka] = -LOST_ON_FLUKA_COLL # Do not kill yet to avoid issues with energy updating

    if np.any(mask_alive):
        # Sanity check
        assert np.all(part.particle_id[idx_alive] == new_pid[mask_alive])
        assert np.all(part.state[idx_alive] == HIT_ON_FLUKA_COLL)

        # Update energy
        E_diff[idx_alive] = part.energy[idx_alive] - data['e'][:npart][mask_alive] * 1.e6
        if np.any(E_diff < -precision):
            raise ValueError(f"FLUKA returned particle with energy higher than incoming particle!")
        E_diff[E_diff < precision] = 0. # Lower cut on energy loss
        part.add_to_energy(-E_diff)
        part.weight[idx_alive]     = data['weight'][:npart][mask_alive]
        coll._acc_ionisation_loss += np.sum(E_diff[idx_alive]*part.weight[idx_alive])

        rpp = part.rpp[idx_alive]  # This is now already updated by the new energy
        part.x[idx_alive]          = data['x'][:npart][mask_alive] / 1000.
        part.px[idx_alive]         = data['xp'][:npart][mask_alive] / rpp / 1000. # This is exact because FLUKA uses director cosine
        part.y[idx_alive]          = data['y'][:npart][mask_alive] / 1000.
        part.py[idx_alive]         = data['yp'][:npart][mask_alive] / rpp / 1000. # This is exact because FLUKA uses director cosine
        part.zeta[idx_alive]       = data['zeta'][:npart][mask_alive] / 1000.
        part.state[idx_alive]      = 1  # These actually survived


    # Add new particles
    # ================
    mask_new = new_pid > max_id
    q_new = data['q'][:npart]
    pdg_id = data['pdg_id'][:npart]
    mask_new &= xc.fluka.engine._mask_particle_return_types(pdg_id, q_new)

    if np.any(mask_new):
        # Check that there is enough room in the particles object
        num_assigned = part._num_lost_particles + part._num_active_particles
        num_free = part._capacity - num_assigned
        num_needed = mask_new.sum()
        if num_free < num_needed:
            raise RuntimeError(f"Too many particles generated by FLUKA ({num_needed} needed, "
                             + f"but only {num_free} free in particles object)!")

        # Parent particle IDs
        parents = new_ppid[mask_new]
        pids = part.particle_id[send_to_fluka]
        max_id_in = int(pids.max())
        assert np.all(parents >= 0)        # Sanity check
        assert parents.max() <= max_id_in  # Sanity check

        # Build dense lookup table: ID -> index in `pids`
        lookup = np.full(max_id_in + 1, -1, dtype=np.int64)
        lookup[pids] = np.arange(pids.size, dtype=np.int64)
        idx_parents = lookup[parents]   # MUCH faster than solution with np.where
        if np.any(idx_parents < 0):
            missing = np.unique(parents[idx_parents < 0])
            raise RuntimeError(f"Parent IDs not found in particle_id: {missing}")

        # Create new particles
        E = data['e'][:npart][mask_new] * 1.e6
        m = data['m'][:npart][mask_new] * 1.e6
        q = data['q'][:npart][mask_new]
        if np.any(E < -precision):
            raise ValueError(f"FLUKA returned particles with negative energy!")
        if np.any(m < -precision):
            raise ValueError(f"FLUKA returned particles with negative mass!")
        # TODO: we set massless particles to have half the energy in their mass.
        # To be adapted when Xsuite can handle neutral and massless particles.
        mask_massless = np.abs(m) < 1.e-12
        mask_neutral = np.abs(q) < 1.e-12
        m[mask_massless | mask_neutral] = E[mask_massless | mask_neutral]/np.sqrt(2)  # Half the energy is put in the mass
        q[mask_massless | mask_neutral] = q0
        delta = np.sqrt(E/m*E/m - 1)/beta0/gamma0 - 1
        rpp   = 1. / (1. + delta)
        new_part = xt.Particles(_context=part._buffer.context,
                p0c = p0c,
                mass0 = m0,
                q0 = q0,
                s = s_in + coll.length + coll.length_front + coll.length_back,
                x = data['x'][:npart][mask_new] / 1000.,
                px = data['xp'][:npart][mask_new] * (1. + delta) / 1000.,
                y = data['y'][:npart][mask_new] / 1000.,
                py = data['yp'][:npart][mask_new] * (1. + delta) / 1000.,
                zeta = data['zeta'][:npart][mask_new] / 1000.,
                delta = delta,
                mass_ratio = m / m0,
                charge_ratio = q / q0,
                at_element = ele_in,
                at_turn = turn_in,
                pdg_id = data['pdg_id'][:npart][mask_new],
                particle_id = new_pid[mask_new],
                parent_particle_id = new_ppid[mask_new],
                weight = data['weight'][:npart][mask_new],
                start_tracking_at_element = start)

        # Correct the deposited energy of parent particles: not everything was lost there.
        E_children = np.bincount(idx_parents, weights=new_part.energy, minlength=part._capacity)
        if np.any(E_children < -precision):
            raise ValueError(f"FLUKA returned particles with summed energy higher than parent particle!")
        # If the parent survived, this should not be done but the energy should be subtracted
        # from the accumulated ionisation loss (as it is accounted for by the child)
        mask_parent_survived = (part.state==1) & (E_children > 0)
        if np.any(mask_parent_survived):
            if np.any(E_diff[mask_parent_survived] - E_children[mask_parent_survived]) < -precision:
                raise ValueError(f"FLUKA returned children with a surviving parent, however, there "
                               + f"was a larger energy loss than the children energy!")
            coll._acc_ionisation_loss -= np.sum(E_children[mask_parent_survived]*part.weight[mask_parent_survived])
            E_children[mask_parent_survived] = 0.
        # If the deposited energy is lower than the rest mass, it cannot be represented by the original
        # particle. We make a virtual particle with fake mass (E=2m) to avoid negative square roots.
        mask_virtual = (part.energy - E_children < part.mass) & (E_children > 0)
        if np.any(mask_virtual):
            virtual_mass = (part.energy - E_children)[mask_virtual] / np.sqrt(2)
            new_ptau = part.ptau.copy()
            old_mass = part.mass[mask_virtual]
            old_ptau = new_ptau[mask_virtual]
            part.chi[mask_virtual] = m0 / virtual_mass * part.charge_ratio[mask_virtual]
            new_ptau[mask_virtual] = (1/beta0 + old_ptau)*old_mass/virtual_mass - 1/beta0
            part.update_ptau(new_ptau)
            part.state[mask_virtual] = -VIRTUAL_ENERGY
        # Now update the parent energies
        part.add_to_energy(-E_children)

        # Add new particles
        new_part._init_random_number_generator()
        # TODO: we instantly kill massless or neutral particles as Xsuite is not ready to handle them.
        new_part.state[mask_massless | mask_neutral] = -MASSLESS_OR_NEUTRAL
        part.add_particles(new_part)
        max_particle_id = new_pid.max()
        if max_particle_id <= xc.fluka.engine.max_particle_id:
            raise ValueError(f"FLUKA returned new particles with IDs {max_particle_id} that are "
                           + f"lower than the highest ID known ({xc.fluka.engine.max_particle_id}).\n"
                           + "This should not happen. Please report this issue to the developers.")
        xc.fluka.engine._max_particle_id = max_particle_id

    # Kill all flagged particles
    part.state[part.state==-LOST_ON_FLUKA_COLL] = LOST_ON_FLUKA_COLL
    part.state[part.state==-VIRTUAL_ENERGY] = VIRTUAL_ENERGY
    # TODO: we instantly kill massless or neutral particles as Xsuite is not ready to handle them.
    part.state[part.state==-MASSLESS_OR_NEUTRAL] = MASSLESS_OR_NEUTRAL

    # Reshuffle
    part.reorganize()

    # Give all particles the same s position to avoid numerical differences
    part.s[part.state > 0] = s_in + coll.length + coll.length_front + coll.length_back



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2024.                 #
# ######################################### #

import io
import numpy as np
from warnings import warn

import xobjects as xo
import xtrack as xt

from .. import json
from ..compare import deep_equal


class ParticleStatsMonitorRecord(xo.Struct):
    count            = xo.Int64[:]
    x_sum1           = xo.Float64[:]
    px_sum1          = xo.Float64[:]
    y_sum1           = xo.Float64[:]
    py_sum1          = xo.Float64[:]
    zeta_sum1        = xo.Float64[:]
    pzeta_sum1       = xo.Float64[:]
    delta_sum1       = xo.Float64[:]
    x_x_sum2         = xo.Float64[:]
    x_px_sum2        = xo.Float64[:]
    x_y_sum2         = xo.Float64[:]
    x_py_sum2        = xo.Float64[:]
    x_zeta_sum2      = xo.Float64[:]
    x_pzeta_sum2     = xo.Float64[:]
    x_delta_sum2     = xo.Float64[:]
    px_px_sum2       = xo.Float64[:]
    px_y_sum2        = xo.Float64[:]
    px_py_sum2       = xo.Float64[:]
    px_zeta_sum2     = xo.Float64[:]
    px_pzeta_sum2    = xo.Float64[:]
    px_delta_sum2    = xo.Float64[:]
    y_y_sum2         = xo.Float64[:]
    y_py_sum2        = xo.Float64[:]
    y_zeta_sum2      = xo.Float64[:]
    y_pzeta_sum2     = xo.Float64[:]
    y_delta_sum2     = xo.Float64[:]
    py_py_sum2       = xo.Float64[:]
    py_zeta_sum2     = xo.Float64[:]
    py_pzeta_sum2    = xo.Float64[:]
    py_delta_sum2    = xo.Float64[:]
    zeta_zeta_sum2   = xo.Float64[:]
    zeta_pzeta_sum2  = xo.Float64[:]
    zeta_delta_sum2  = xo.Float64[:]
    pzeta_pzeta_sum2 = xo.Float64[:]
    pzeta_delta_sum2 = xo.Float64[:]
    delta_delta_sum2 = xo.Float64[:]

_RECORD_FIELD_NAMES = {f.name for f in ParticleStatsMonitorRecord._fields}

class ParticleStatsMonitor(xt.BeamElement):
    _xofields = {
        'part_id_start':      xo.Int64,
        'part_id_end':        xo.Int64,
        'start_at_turn':      xo.Int64,
        'stop_at_turn':       xo.Int64,
        'frev':               xo.Float64,
        'sampling_frequency': xo.Float64,
        'cached':             xo.Int8[:],
        '_index':             xt.RecordIndex,
        'data':               ParticleStatsMonitorRecord,
        '_selector':          xo.Int16
    }

    _isthick = False
    needs_rng = False
    prototype = "monitor"
    allow_track = True
    iscollective = False
    has_backtrack = False
    behaves_like_drift = True
    allow_rot_and_shift = True
    allow_loss_refinement = False
    name_associated_aperture = None
    skip_in_loss_location_refinement = True

    _noexpr_fields = {'name', 'line'}
    _extra_c_sources = [
        '#include "xcoll/beam_elements/elements_src/monitor.h"'
    ]
    _store_in_to_dict = ["beta0", "gamma0", "mass0"]

    def __init__(self, **kwargs):
        """Monitor to save particle statistics (mean and variance of x,
        px, y, py, zeta, pzeta, delta) of (potentially a subset of)
        particles over a range of turns.

        The monitor allows for arbitrary sampling rate and can thus not
        only be used to monitor bunch positions, but also to record
        schottky spectra. Internally, the particle arrival time is used
        when determining the record index:

            i = sampling_frequency * (
                        ( at_turn - start_turn ) / f_rev
                        - zeta / beta0 / c0
                )

        where zeta=(s-beta0*c0*t) is the longitudinal coordinate of the
        particle, beta0 the relativistic beta factor of the particle, c0
        is the speed of light, at_turn is the current turn number, f_rev
        is the revolution frequency, and sampling_frequency is the
        sampling frequency.

        Note that the index is rounded, i.e. the result array represents
        data of particles equally distributed around the reference
        particle. For example, if the sampling_frequency is twice the
        revolution frequency, the first item contains data from
        particles in the range zeta/circumference = -0.25 .. 0.25, the
        second item in the range 0.25 .. 0.75 and so on.

        Args:
            num_particles (int, optional):
                Number of particles to monitor, starting from 0.
                Defaults to -1 which means ALL.
            particle_id_range (tuple, optional):
                Range of particle ids to monitor (start, stop). Stop is
                exclusive. Defaults to (particle_id_start,
                particle_id_start+num_particles).
            start_at_turn (int):
                First turn of reference particle (inclusive) at which to
                monitor. Defaults to 0.
            stop_at_turn (int):
                Last turn of reference particle (exclusive) at which to
                monitor. Defaults to start_at_turn + 1.
            frev (float):
                Revolution frequency in Hz of circulating beam (used to
                relate turn number to sample index). Defaults to 1.
            sampling_frequency (float):
                Sampling frequency in Hz. Defaults to 1.
            mean (bool):
                Whether or not to monitor means. Defaults to True.
            variance (bool):
                Whether or not to monitor variances. Defaults to True,
                unless mean is explicitly set to True.
            columns (list of str):
                List of variables to monitor. Can include any of 'x',
                'px', 'y', 'py', 'zeta', 'pzeta', 'delta'. Defaults to
                empty list, in which case all variables are monitored.
        """

        to_assign = {}
        if '_xobject' not in kwargs:
            # Get all flags related to what to monitor
            mon_mean = kwargs.pop('mean', None)
            mon_variance = kwargs.pop('variance', None)
            if mon_mean is None and mon_variance is None:
                mon_mean = mon_variance = True
            elif mon_mean is None:
                mon_mean = True
            elif mon_variance is None:
                mon_variance = not mon_mean
            columns = kwargs.pop('columns', [])
            mon_x = 'x' in columns
            mon_px = 'px' in columns
            mon_y = 'y' in columns
            mon_py = 'py' in columns
            mon_zeta = 'zeta' in columns
            mon_pzeta = 'pzeta' in columns
            mon_delta = 'delta' in columns
            if all(v is False for v in [mon_x, mon_px, mon_y, mon_py, mon_zeta,
                                        mon_pzeta, mon_delta]):
                mon_x = True
                mon_px = True
                mon_y = True
                mon_py = True
                mon_zeta = True
                mon_pzeta = True
                mon_delta = True
            kwargs['_selector']  =     int(mon_x)
            kwargs['_selector'] +=   2*int(mon_px)
            kwargs['_selector'] +=   4*int(mon_y)
            kwargs['_selector'] +=   8*int(mon_py)
            kwargs['_selector'] +=  16*int(mon_zeta)
            kwargs['_selector'] +=  32*int(mon_pzeta)
            kwargs['_selector'] +=  64*int(mon_delta)
            kwargs['_selector'] += 128*int(mon_mean)
            kwargs['_selector'] += 256*int(mon_variance)

            # Get ranges
            if 'particle_id_range' in kwargs:
                assert 'num_particles' not in kwargs
                particle_id_range = kwargs.pop('particle_id_range')
                kwargs['part_id_start'] = int(particle_id_range[0])
                kwargs['part_id_end'] = int(particle_id_range[1])
            else:
                num_particles = int(kwargs.pop('num_particles', -1))
                if num_particles == -1:
                    kwargs['part_id_start'] = 0
                    kwargs['part_id_end'] = -1
                else:
                    kwargs['part_id_start'] = kwargs.pop(
                                                'particle_id_start', 0)
                    kwargs['part_id_end'] = kwargs['part_id_start'] \
                                            + num_particles
            start = int(kwargs.get('start_at_turn', 0))
            stop = int(kwargs.get('stop_at_turn', start + 1))
            frev = kwargs.get('frev', 1.)
            samp = kwargs.get('sampling_frequency', 1.)
            kwargs['start_at_turn'] = start
            kwargs['stop_at_turn'] = stop
            kwargs['frev'] = frev
            kwargs['sampling_frequency'] = samp

            # Prepare the arrays. Explicitely init with zeros (instead of size
            # only) to have consistent initial values
            if "data" not in kwargs:
                dd = {}
                size = int(round((stop-start) * samp / frev))
                for ff in ParticleStatsMonitorRecord._fields:
                    if ff.name == 'count':
                        dd[ff.name] = np.zeros(size, dtype=np.int64)
                    elif ff.name.endswith('_sum1'):
                        if not mon_mean:
                            dd[ff.name] = np.zeros(1, dtype=np.float64)
                        elif (ff.name.startswith('x_') and mon_x) or \
                             (ff.name.startswith('px_') and mon_px) or \
                             (ff.name.startswith('y_') and mon_y) or \
                             (ff.name.startswith('py_') and mon_py) or \
                             (ff.name.startswith('zeta_') and mon_zeta) or \
                             (ff.name.startswith('pzeta_') and mon_pzeta) or \
                             (ff.name.startswith('delta_') and mon_delta):
                            dd[ff.name] = np.zeros(size, dtype=np.float64)
                        else:
                            dd[ff.name] = np.zeros(1, dtype=np.float64)
                    elif ff.name.endswith('_sum2'):
                        if not mon_variance:
                            dd[ff.name] = np.zeros(1, dtype=np.float64)
                        else:
                            coords = ff.name[:-5].split('_')
                            if len(coords) != 2:
                                raise ValueError(f"Unexpected field {ff.name} "
                                              "in ParticleStatsMonitorRecord!")
                            if ('x' in coords and not mon_x) or \
                               ('px' in coords and not mon_px) or \
                               ('y' in coords and not mon_y) or \
                               ('py' in coords and not mon_py) or \
                               ('zeta' in coords and not mon_zeta) or \
                               ('pzeta' in coords and not mon_pzeta) or \
                               ('delta' in coords and not mon_delta):
                                dd[ff.name] = np.zeros(1, dtype=np.float64)
                            else:
                                dd[ff.name] = np.zeros(size, dtype=np.float64)
                    else:
                        raise ValueError(f"Unknown field {ff.name} in "
                                         "ParticleStatsMonitorRecord!")
                    kwargs['data'] = dd
            kwargs.setdefault('cached', np.zeros(len(kwargs['data']['count']),
                                                 dtype=np.int8))
            to_assign['_beta0'] = kwargs.pop('beta0', None)
            to_assign['_gamma0'] = kwargs.pop('gamma0', None)
            to_assign['_mass0'] = kwargs.pop('mass0', None)
        super().__init__(**kwargs)
        for kk, vv in to_assign.items():
            setattr(self, kk, vv)

    def to_json(self, file, indent=2, backend=['orjson', 'msgspec', 'json']):
        json.dump(self.to_dict(), file, indent=indent, backend=backend)

    @classmethod
    def from_json(cls, file):
        if not hasattr(file, '__iter__') or isinstance(file, (str, io.IOBase)):
            file = [file]
        dct = {}
        data = {}
        for f in file:
            this_dct = json.load(f)
            this_data = this_dct.pop('data')
            if dct == {}:
                dct = this_dct
                data = {kk: np.array(vv) for kk, vv in this_data.items()}
            else:
                if not deep_equal(this_dct, dct):
                    raise ValueError(f"JSON file {f} not compatible with "
                                      "previous ones!")
                for key, value in this_data.items():
                    if key not in data:
                        raise ValueError(f"JSON file {f} not compatible with "
                                          "previous ones!")
                    data[key] += np.array(value)
        self = cls.from_dict(dct | {'data': data})
        return self

    @classmethod
    def install(cls, line, name, *, at=None, s_tol=1.e-6, at_s=None, **kwargs):
        """Install a monitor in the line. The monitor will be configured
        with the line's reference particle parameters.
        """
        if at_s is not None:
            warn("Warning: `at_s` is deprecated and will be removed in "
                 "the future. Please use `at` instead.", FutureWarning)
            at = at_s
        if name in line.element_names:
            raise ValueError(f"Element {name} already exists in the line as "
                             f"{line[name].__class__.__name__}.")
        self = cls(**kwargs)
        line.insert(name, self, at=at, s_tol=s_tol)
        self.configure(line)
        return self

    def configure(self, line=None, *, beta0=None, gamma0=None, mass0=None):
        """Set optics parameters from line's reference particle. If
        `line` is not provided, `beta0`, `gamma0` and `mass0` must be
        provided explicitly.
        """
        if beta0 is not None and gamma0 is not None and mass0 is not None:
            self._beta0 = beta0
            self._gamma0 = gamma0
            self._mass0 = mass0
        elif line is not None and hasattr(line, 'particle_ref') \
        and line.particle_ref is not None:
            self._beta0 = line.particle_ref.beta0[0]
            self._gamma0 = line.particle_ref.gamma0[0]
            self._mass0 = line.particle_ref.mass0
        else:
            raise ValueError("Either a line with a particle_ref, or beta0, "
                             "gamma0 and mass0 must be provided!")

    def reset(self):
        """Reset the monitor data (to avoid unwanted accumulation)."""
        for field in [f.name for f in ParticleStatsMonitorRecord._fields]:
            ff = getattr(self.data, field)
            zeros = np.zeros(len(ff), dtype=ff._itemtype._dtype)
            setattr(self.data, field, zeros)
        for i in np.arange(len(self.count)):
            self.cached[i] = 0

    @property
    def beta0(self):
        return self._beta0

    @property
    def gamma0(self):
        return self._gamma0

    @property
    def mass0(self):
        return self._mass0

    @property
    def monitor_x(self):
        return bool(self._selector % 2)

    @property
    def monitor_px(self):
        return bool((self._selector >> 1) % 2)

    @property
    def monitor_y(self):
        return bool((self._selector >> 2) % 2)

    @property
    def monitor_py(self):
        return bool((self._selector >> 3) % 2)

    @property
    def monitor_zeta(self):
        return bool((self._selector >> 4) % 2)

    @property
    def monitor_pzeta(self):
        return bool((self._selector >> 5) % 2)

    @property
    def monitor_delta(self):
        return bool((self._selector >> 6) % 2)

    @property
    def monitor_mean(self):
        return bool((self._selector >> 7) % 2)

    @property
    def monitor_variance(self):
        return bool((self._selector >> 8) % 2)

    @property
    def turns(self):
        self._calculate()
        return self._turns

    @property
    def pc_mean(self):
        if self.monitor_delta and self.monitor_mean:
            self._calculate()
            one_plus_delta = 1 + self.delta_mean
            return one_plus_delta * self.beta0 * self.gamma0 * self.mass0
        else:
            raise ValueError("Momentum mean not available! Set "
                             "monitor_delta=True and monitor_mean=True.")

    @property
    def pc_var(self):
        if self.monitor_delta and self.monitor_variance:
            self._calculate()
            var_delta = self.delta_var
            return var_delta * self.beta0**2 * self.gamma0**2 * self.mass0**2
        else:
            raise ValueError("Momentum variance not available! Set "
                             "monitor_delta=True and monitor_variance=True.")

    @property
    def energy_mean(self):
        if self.monitor_pzeta and self.monitor_mean:
            self._calculate()
            one_plus_pzeta = 1 + self.beta0**2 * self.pzeta_mean
            return one_plus_pzeta * self.gamma0 * self.mass0
        else:
            raise ValueError("Energy mean not available! Set "
                             "monitor_pzeta=True and monitor_mean=True.")

    @property
    def energy_var(self):
        if self.monitor_pzeta and self.monitor_variance:
            self._calculate()
            var_pzeta = self.beta0**4 * self.pzeta_mean**2
            return var_pzeta * self.gamma0**2 * self.mass0**2
        else:
            raise ValueError("Energy variance not available! Set "
                             "monitor_pzeta=True and monitor_variance=True.")


    def _set_arr(self, name, mask, value):
        # Assign to attribute (create if does not exist yet):
        arr = getattr(self, name, None)
        if arr is None:
            arr = np.full(self.count.shape, np.nan, dtype=float)
            setattr(self, name, arr)
        arr[mask] = value

    def _calculate(self):
        N = self.count
        mask = (N > 0) & (self.cached == 0)
        if not np.any(mask):
            return

        # Calculate mean, variance, and std
        N = N[mask]
        self._turns = np.array(range(self.start_at_turn,
                                     self.stop_at_turn))[mask]

        # NaN for zero particles is expected behaviour
        with np.errstate(invalid='ignore'):
            for field in [f.name for f in ParticleStatsMonitorRecord._fields]:
                if field.endswith('_sum1'):
                    x = field[:-5]
                    ff = getattr(self, field)
                    ff = ff[mask] if len(ff) == len(mask) else ff
                    mean = ff / N
                    self._set_arr(f'_{x}_mean', mask, mean)
            for field in [f.name for f in ParticleStatsMonitorRecord._fields]:
                if field.endswith('_sum2'):
                    x1, x2 = field[:-5].split('_')
                    mean1 = getattr(self, f'_{x1}_mean')
                    mean2 = getattr(self, f'_{x2}_mean')
                    ff = getattr(self, field)
                    ff = ff[mask] if len(ff) == len(mask) else ff
                    variance = ff / (N - 1) - mean1 * mean2 * N / (N - 1)
                    self._set_arr(f'_{x1}_{x2}_var', mask, variance)
        for i in np.arange(len(self.count))[mask]:
            self.cached[i] = 1

    def __getattr__(self, attr):
        if attr in _RECORD_FIELD_NAMES:
            return getattr(self.data, attr).to_nparray()

        if attr.startswith('_'):
            raise AttributeError(f"Attribute {attr} not set!")

        if attr.endswith('_var') or attr.endswith('_mean'):
            self._calculate()
            return getattr(self, f'_{attr}')

        return super().__getattribute__(attr)


class EmittanceMonitor(ParticleStatsMonitor):
    _xofields = ParticleStatsMonitor._xofields
    _depends_on = [ParticleStatsMonitor]
    _extra_c_sources = [
        '#include "xcoll/beam_elements/elements_src/emittance_monitor.h"'
    ]

    def __init__(self, **kwargs):
        """
        Monitor to save the normalised beam emittance

        Similar to the BeamSizeMonitor and BeamPositionMonitor, it
        allows for arbitrary sampling rate and can thus not only be used
        to monitor bunch emittance, but also to record coasting beams.
        See their documentation for more information on how to use
        `frev` and `sampling_frequency`.

        Args:
            num_particles (int, optional): Number of particles to
                monitor, starting from 0. Defaults to -1 which means
                ALL.
            particle_id_range (tuple, optional): Range of particle ids
                to monitor (start, stop). Stop is exclusive. Defaults to
                (particle_id_start, particle_id_start+num_particles).
            start_at_turn (int): First turn of reference particle
                (inclusive) at which to monitor. Defaults to 0.
            stop_at_turn (int): Last turn of reference particle
                (exclusive) at which to monitor. Defaults to
                start_at_turn + 1.
            frev (float): Revolution frequency in Hz of circulating beam
                (used to relate turn number to sample index). Defaults
                to 1.
            sampling_frequency (float): Sampling frequency in Hz.
                Defaults to 1.
            horizontal (bool): Whether or not to monitor the horizontal
                plane. Defaults to True.
            vertical (bool): Whether or not to monitor the vertical
                plane. Defaults to True.
            longitudinal (bool): Whether or not to monitor the
                longitudinal plane. Defaults to True.
        """

        to_assign = {}
        if '_xobject' not in kwargs:
            kwargs['mean'] = True
            kwargs['variance'] = True
            horizontal = kwargs.pop('horizontal', None)
            vertical = kwargs.pop('vertical', None)
            longitudinal = kwargs.pop('longitudinal', None)
            # Small hack to allow e.g. horizontal=False to imply
            # vertical=True and longitudinal=True
            setval = np.unique([v for v in [horizontal, vertical, longitudinal]
                                if v is not None])
            if len(setval) == 0:
                setval = True
            elif len(setval) == 1:
                setval = not setval[0]
            else:
                setval = False
            if horizontal is None: horizontal = setval
            if vertical is None: vertical = setval
            if longitudinal is None: longitudinal = setval
            kwargs['columns'] = []
            if horizontal:
                kwargs['columns'] += ['x', 'px']
            if vertical:
                kwargs['columns'] += ['y', 'py']
            if longitudinal:
                kwargs['columns'] += ['zeta', 'pzeta']
            if kwargs.pop('monitor_delta', False):
                kwargs['columns'] += ['delta']
            to_assign['_suppress_warnings'] = kwargs.pop('suppress_warnings',
                                                         False)
        super().__init__(**kwargs)
        for kk, vv in to_assign.items():
            setattr(self, kk, vv)
        if not hasattr(self, 'cached_modes'):
            self.cached_modes = np.zeros(len(self.cached), dtype=np.int8)

    def reset(self):
        super().reset()
        self.cached_modes[:] = 0

    @property
    def suppress_warnings(self):
        return self._suppress_warnings

    @suppress_warnings.setter
    def suppress_warnings(self, val):
        self._suppress_warnings = bool(val)

    @property
    def gemitt_x(self):
        self._check_horizontal()
        self._calculate()
        return self._gemitt_x

    @property
    def gemitt_y(self):
        self._check_vertical()
        self._calculate()
        return self._gemitt_y

    @property
    def gemitt_zeta(self):
        self._check_longitudinal()
        self._calculate()
        return self._gemitt_zeta

    @property
    def nemitt_x(self):
        self._check_horizontal()
        self._calculate()
        return self._gemitt_x * self.beta0 * self.gamma0

    @property
    def nemitt_y(self):
        self._check_vertical()
        self._calculate()
        return self._gemitt_y * self.beta0 * self.gamma0

    @property
    def nemitt_zeta(self):
        self._check_longitudinal()
        self._calculate()
        return self._gemitt_zeta * self.beta0 * self.gamma0

    @property
    def gemitt_I(self):
        self._check_horizontal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_I

    @property
    def gemitt_II(self):
        self._check_vertical()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_II

    @property
    def gemitt_III(self):
        self._check_longitudinal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_III

    @property
    def nemitt_I(self):
        self._check_horizontal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_I * self.beta0 * self.gamma0

    @property
    def nemitt_II(self):
        self._check_vertical()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_II * self.beta0 * self.gamma0

    @property
    def nemitt_III(self):
        self._check_longitudinal()
        self._calculate()
        self._calculate_modes()
        return self._gemitt_III * self.beta0 * self.gamma0


    @property
    def horizontal(self):
        return self.monitor_x and self.monitor_px

    @property
    def vertical(self):
        return self.monitor_y and self.monitor_py

    @property
    def longitudinal(self):
        return self.monitor_zeta and self.monitor_pzeta


    def _check_horizontal(self):
        if not self.horizontal:
            raise ValueError("Horizontal plane not monitored!")

    def _check_vertical(self):
        if not self.vertical:
            raise ValueError("Vertical plane not monitored!")

    def _check_longitudinal(self):
        if not self.longitudinal:
            raise ValueError("Longitudinal plane not monitored!")


    def _calculate(self):
        N = self.count
        mask = (N > 0) & (self.cached == 0)
        if not np.any(mask):
            return

        super()._calculate()
        self.cached_modes[:] = 0

        # Calculate emittances
        gemitt_x = np.sqrt(self.x_x_var * self.px_px_var - self.x_px_var**2)
        gemitt_y = np.sqrt(self.y_y_var * self.py_py_var - self.y_py_var**2)
        gemitt_zeta = np.sqrt(self.zeta_zeta_var * self.pzeta_pzeta_var
                              - self.zeta_pzeta_var**2)
        setattr(self, '_gemitt_x', gemitt_x)
        setattr(self, '_gemitt_y', gemitt_y)
        setattr(self, '_gemitt_zeta', gemitt_zeta)


    def _calculate_modes(self):
        # Calculate emittance modes
        N = self.count
        mask = (N > 0) & (self.cached_modes == 0)
        if not np.any(mask):
            return

        S = np.array([[ 0., 1., 0., 0., 0., 0.],
                      [-1., 0., 0., 0., 0., 0.],
                      [ 0., 0., 0., 1., 0., 0.],
                      [ 0., 0.,-1., 0., 0., 0.],
                      [ 0., 0., 0., 0., 0., 1.],
                      [ 0., 0., 0., 0.,-1., 0.]])

        # Create arrays for modes in attributes if needed
        for nn in ['gemitt_I', 'gemitt_II', 'gemitt_III']:
            if getattr(self, f'_{nn}', None) is None:
                arr = np.full(N.shape, -1, dtype=float)
                setattr(self, f'_{nn}', arr)

        # Loop over turns
        for i in range(len(N)):
            if not mask[i]:
                continue

            if N[i] < 25:
                # Not enough statistics for a reliable calculation of the modes
                self._gemitt_I[i] = 0
                self._gemitt_II[i] = 0
                self._gemitt_III[i] = 0
                continue

            if self.horizontal:
                block_x = np.array([[self.x_x_var[i],  self.x_px_var[i]],
                                    [self.x_px_var[i], self.px_px_var[i]]])
            else:
                block_x = np.zeros((2, 2))
            if self.vertical:
                block_y = np.array([[self.y_y_var[i],  self.y_py_var[i]],
                                    [self.y_py_var[i], self.py_py_var[i]]])
            else:
                block_y = np.zeros((2, 2))
            if self.longitudinal:
                block_z = np.array([[self.zeta_zeta_var[i],
                                     self.zeta_pzeta_var[i]],
                                    [self.zeta_pzeta_var[i],
                                     self.pzeta_pzeta_var[i]]])
            else:
                block_z = np.zeros((2, 2))
            if self.horizontal and self.vertical:
                block_xy = np.array([[self.x_y_var[i],  self.x_py_var[i]],
                                     [self.px_y_var[i], self.px_py_var[i]]])
            else:
                block_xy = np.zeros((2, 2))
            if self.horizontal and self.longitudinal:
                block_xz = np.array([[self.x_zeta_var[i],
                                      self.x_pzeta_var[i]],
                                     [self.px_zeta_var[i],
                                      self.px_pzeta_var[i]]])
            else:
                block_xz = np.zeros((2, 2))
            if self.vertical and self.longitudinal:
                block_yz = np.array([[self.y_zeta_var[i],
                                      self.y_pzeta_var[i]],
                                     [self.py_zeta_var[i],
                                      self.py_pzeta_var[i]]])
            else:
                block_yz = np.zeros((2, 2))

            covariance_S = np.dot(
                        np.block([[block_x,    block_xy,   block_xz],
                                  [block_xy.T, block_y,    block_yz],
                                  [block_xz.T, block_yz.T, block_z]]),
                        S)

            # Check for all zero matrix -> zero emittance
            if np.all(covariance_S < 1E-16):
                self._gemitt_I[i] = 0
                self._gemitt_II[i] = 0
                self._gemitt_III[i] = 0
                continue

            cond_number = np.linalg.cond(covariance_S)
            if cond_number > 1e10 and not self.suppress_warnings:
                print(f"Warning: High condition number at time step {i}: "
                    + f"{cond_number}.\n{N[i]} particles logged.")

            rank = np.linalg.matrix_rank(covariance_S)
            expected_rank = int(self.horizontal) + int(self.vertical) + int(self.longitudinal)
            if rank < expected_rank and not self.suppress_warnings:
                print(f"Warning: Matrix is rank deficient at time step {i}: "
                    + f"rank {rank} instead of expected {len(covariance_S)}.\n"
                    + f"{N[i]} particles logged.")

            from xtrack.linear_normal_form import compute_linear_normal_form
            _, _, _, eigenvalues = compute_linear_normal_form(covariance_S)
            self._gemitt_I[i] = eigenvalues[0].imag
            self._gemitt_II[i] = eigenvalues[1].imag
            self._gemitt_III[i] = eigenvalues[2].imag

        # Set calculated modes as cached
        self.cached_modes[mask] = 1



# copyright ############################### #
# This file is part of the Xcoll package.   #
# Copyright (c) CERN, 2025.                 #
# ######################################### #

from .element_types import DEFAULT_XCOLL_ELEMENTS, EXTRA_XCOLL_ELEMENTS
from .element_inits import XCOLL_ELEMENTS_INIT_DEFAULTS
